from autokeras.pretrained.text_classifier import SentimentAnalysis

def test_sentiment_analysis():
    sentiment_analyzer = SentimentAnalysis()

    positive_polarity = sentiment_analyzer.predict("The model is working really well.")
    if positive_polarity <= 0.5:
        raise AssertionError()

    negative_polarity = sentiment_analyzer.predict("The university intake has reduced drastically this year.")
    if negative_polarity >= 0.5:
        raise AssertionError()



from random import randrange

from autokeras.bayesian import SearchTree, contain
from autokeras.net_transformer import transform
from autokeras.search import Searcher


class RandomSearcher(Searcher):
    """ Class to search for neural architectures using Random search strategy.
    Attributes:
        search_tree: The network morphism search tree
    """

    def __init__(self, n_output_node, input_shape, path, metric, loss, generators, verbose,
                 trainer_args=None,
                 default_model_len=None,
                 default_model_width=None):
        super(RandomSearcher, self).__init__(n_output_node, input_shape,
                                             path, metric, loss, generators,
                                             verbose, trainer_args, default_model_len,
                                             default_model_width)
        self.search_tree = SearchTree()

    def generate(self, multiprocessing_queue):
        """Generate the next neural architecture.

        Args:
            multiprocessing_queue: the Queue for multiprocessing return value.

        Returns:
            list of 2-element tuples: generated_graph and other_info,
            for random searcher the length of list is 1.
            generated_graph: An instance of Graph.
            other_info: Anything to be saved in the training queue together with the architecture.

        """
        random_index = randrange(len(self.history))
        model_id = self.history[random_index]['model_id']
        graph = self.load_model_by_id(model_id)
        new_father_id = None
        generated_graph = None
        for temp_graph in transform(graph):
            if not contain(self.descriptors, temp_graph.extract_descriptor()):
                new_father_id = model_id
                generated_graph = temp_graph
                break
        if new_father_id is None:
            new_father_id = 0
            generated_graph = self.generators[0](self.n_classes, self.input_shape). \
                generate(self.default_model_len, self.default_model_width)

        return [(generated_graph, new_father_id)]

    def update(self, other_info, model_id, *args):
        """ Update the controller with evaluation result of a neural architecture.

        Args:
            other_info: Anything. In our case it is the father ID in the search tree.
            model_id: An integer.
        """
        father_id = other_info
        self.search_tree.add_child(father_id, model_id)


# coding: utf-8

import math
import torch
from torch import nn
from torch.nn import functional as F

from .modules import conv1d, conv_transpose1d, embedding, linear
from .modules import SinusoidalEncoding, Conv1dGLU


class Encoder(nn.Module):
    def __init__(self, n_vocab, embed_dim, n_speakers, speaker_embed_dim, padding_idx=None, embedding_weight_std=0.1,
                 convolutions=((64, 5, .1),) * 7, dropout=0.1, apply_grad_scaling=False):
        super(Encoder, self).__init__()
        self.dropout = dropout
        self.num_attention_layers = None
        self.apply_grad_scaling = apply_grad_scaling

        # Text input embeddings
        self.embed_tokens = embedding(
            n_vocab, embed_dim, padding_idx, embedding_weight_std)

        self.n_speakers = n_speakers

        # Non causual convolution blocks
        in_channels = embed_dim
        self.convolutions = nn.ModuleList()
        std_mul = 1.0
        for (out_channels, kernel_size, dilation) in convolutions:
            if in_channels != out_channels:
                # Conv1d + ReLU
                self.convolutions.append(
                    conv1d(in_channels, out_channels, kernel_size=1, padding=0,
                           dilation=1, std_mul=std_mul))
                self.convolutions.append(nn.ReLU(inplace=True))
                in_channels = out_channels
                std_mul = 2.0
            self.convolutions.append(
                Conv1dGLU(n_speakers, speaker_embed_dim,
                          in_channels, out_channels, kernel_size, causal=False,
                          dilation=dilation, dropout=dropout, std_mul=std_mul,
                          residual=True))
            in_channels = out_channels
            std_mul = 4.0
        # Last 1x1 convolution
        self.convolutions.append(conv1d(in_channels, embed_dim, kernel_size=1,
                                        padding=0, dilation=1, std_mul=std_mul,
                                        dropout=dropout))

    def forward(self, text_sequences, text_positions=None, lengths=None,
                speaker_embed=None):
        if self.n_speakers != 1 and speaker_embed is None:
            raise AssertionError("Expected \033[1;31m<self.n_speakers>\033[m to be 1 or \033[1:31m<speaker_embed>[m to be not None, but was not")
        # embed text_sequences
        x = self.embed_tokens(text_sequences.long())
        x = F.dropout(x, p=self.dropout, training=self.training)

        # expand speaker embedding for all time steps
        speaker_embed_btc = None

        input_embedding = x

        # B x T x C -> B x C x T
        x = x.transpose(1, 2)

        # １D conv blocks
        for f in self.convolutions:
            x = f(x, speaker_embed_btc) if isinstance(f, Conv1dGLU) else f(x)

        # Back to B x T x C
        keys = x.transpose(1, 2)

        # scale gradients (this only affects backward, not forward)
        # add output to input embedding for attention
        values = (keys + input_embedding) * math.sqrt(0.5)

        return keys, values


class AttentionLayer(nn.Module):
    def __init__(self, conv_channels, embed_dim, dropout=0.1, window_ahead=3, window_backward=1):
        super(AttentionLayer, self).__init__()
        self.query_projection = linear(conv_channels, embed_dim)
        self.key_projection = None
        self.value_projection = None
        self.out_projection = linear(embed_dim, conv_channels)
        self.dropout = dropout
        self.window_ahead = window_ahead
        self.window_backward = window_backward

    def forward(self, query, encoder_out, mask=None, last_attended=None):
        keys, values = encoder_out
        residual = query

        # attention
        x = self.query_projection(query)
        x = torch.bmm(x, keys)

        mask_value = -float("inf")

        if last_attended is not None:
            backward = last_attended - self.window_backward
            if backward > 0:
                x[:, :, :backward] = mask_value
            ahead = last_attended + self.window_ahead
            if ahead < x.size(-1):
                x[:, :, ahead:] = mask_value

        # softmax over last dim
        # (B, tgt_len, src_len)
        sz = x.size()
        x = F.softmax(x.view(sz[0] * sz[1], sz[2]), dim=1)
        x = x.view(sz)
        attn_scores = x

        x = F.dropout(x, p=self.dropout, training=self.training)

        x = torch.bmm(x, values)

        # scale attention output
        s = values.size(1)
        x = x * (s * math.sqrt(1.0 / s))

        # project back
        x = self.out_projection(x)
        x = (x + residual) * math.sqrt(0.5)
        return x, attn_scores


class Decoder(nn.Module):
    def __init__(self, embed_dim, n_speakers, speaker_embed_dim, in_dim=80, r=5, max_positions=512,
                 preattention=((128, 5, 1),) * 4, convolutions=((128, 5, 1),) * 4, attention=True, dropout=0.1,
                 use_memory_mask=False, force_monotonic_attention=False, query_position_rate=1.0,
                 key_position_rate=1.29, window_ahead=3, window_backward=1):
        super(Decoder, self).__init__()
        self.dropout = dropout
        self.in_dim = in_dim
        self.r = r
        self.query_position_rate = query_position_rate
        self.key_position_rate = key_position_rate

        # Position encodings for query (decoder states) and keys (encoder states)
        self.embed_query_positions = SinusoidalEncoding(
            max_positions, convolutions[0][0])
        self.embed_keys_positions = SinusoidalEncoding(
            max_positions, embed_dim)
        # Used for compute multiplier for positional encodings
        self.speaker_proj1, self.speaker_proj2 = None, None

        # Prenet: causal convolution blocks
        self.preattention = nn.ModuleList()
        in_channels = in_dim * r
        std_mul = 1.0
        for out_channels, kernel_size, dilation in preattention:
            if in_channels != out_channels:
                # Conv1d + ReLU
                self.preattention.append(
                    conv1d(in_channels, out_channels, kernel_size=1, padding=0,
                           dilation=1, std_mul=std_mul))
                self.preattention.append(nn.ReLU(inplace=True))
                in_channels = out_channels
                std_mul = 2.0
            self.preattention.append(
                Conv1dGLU(n_speakers, speaker_embed_dim,
                          in_channels, out_channels, kernel_size, causal=True,
                          dilation=dilation, dropout=dropout, std_mul=std_mul,
                          residual=True))
            in_channels = out_channels
            std_mul = 4.0

        # Causal convolution blocks + attention layers
        self.convolutions = nn.ModuleList()
        self.attention = nn.ModuleList()

        for i, (out_channels, kernel_size, dilation) in enumerate(convolutions):
            if in_channels != out_channels:
                raise AssertionError("Expected \033[1;31m<in_channels>\033[m to be equal to \033[1:31m<out_channels>[m, but was not")
            self.convolutions.append(
                Conv1dGLU(n_speakers, speaker_embed_dim,
                          in_channels, out_channels, kernel_size, causal=True,
                          dilation=dilation, dropout=dropout, std_mul=std_mul,
                          residual=False))
            self.attention.append(
                AttentionLayer(out_channels, embed_dim, dropout=dropout, window_ahead=window_ahead,
                               window_backward=window_backward)
                if attention[i] else None)
            in_channels = out_channels
            std_mul = 4.0
        # Last 1x1 convolution
        self.last_conv = conv1d(in_channels, in_dim * r, kernel_size=1,
                                padding=0, dilation=1, std_mul=std_mul,
                                dropout=dropout)

        # Mel-spectrogram (before sigmoid) -> Done binary flag
        self.fc = linear(in_dim * r, 1)

        self.max_decoder_steps = 200
        self.min_decoder_steps = 10
        self.use_memory_mask = use_memory_mask
        self.force_monotonic_attention = [force_monotonic_attention] * len(convolutions)

    def forward(self, encoder_out, inputs=None,
                text_positions=None, frame_positions=None,
                speaker_embed=None, lengths=None):
        if inputs is None:
            if text_positions is None:
                raise AssertionError("Expected \033[1;31m<text_positions>\033[m to be not None, but was")
            self.start_fresh_sequence()
            outputs = self.incremental_forward(encoder_out, text_positions)
            return outputs

        # Grouping multiple frames if necessary

    def incremental_forward(self, encoder_out, text_positions, initial_input=None, test_inputs=None):
        keys, values = encoder_out
        b = keys.size(0)

        # position encodings
        w = self.key_position_rate
        text_pos_embed = self.embed_keys_positions(text_positions, w)
        keys = keys + text_pos_embed

        # transpose only once to speed up attention layers
        keys = keys.transpose(1, 2).contiguous()

        decoder_states = []
        outputs = []
        alignments = []
        dones = []
        # intially set to zeros
        last_attended = [None] * len(self.attention)
        for idx, v in enumerate(self.force_monotonic_attention):
            last_attended[idx] = 0 if v else None

        num_attention_layers = sum([layer is not None for layer in self.attention])
        t = 0
        if initial_input is None:
            initial_input = keys.data.new(b, 1, self.in_dim * self.r).zero_()
        current_input = initial_input
        while True:
            # frame pos start with 1.
            frame_pos = keys.data.new(b, 1).fill_(t + 1).long()
            w = self.query_position_rate
            frame_pos_embed = self.embed_query_positions(frame_pos, w)

            if t > 0:
                current_input = outputs[-1]
            output_tensor = current_input
            output_tensor = F.dropout(output_tensor, p=self.dropout, training=self.training)

            # Prenet
            for f in self.preattention:
                if isinstance(f, Conv1dGLU):
                    output_tensor = f.incremental_forward(output_tensor)
                else:
                    try:
                        output_tensor = f.incremental_forward(output_tensor, )
                    except AttributeError:
                        output_tensor = f(output_tensor)

            # Casual convolutions + Multi-hop attentions
            ave_alignment = None
            for idx, (f, attention) in enumerate(zip(self.convolutions,
                                                     self.attention)):
                residual = output_tensor
                if isinstance(f, Conv1dGLU):
                    output_tensor = f.incremental_forward(output_tensor)

                if attention is not None:
                    assert isinstance(f, Conv1dGLU)
                    output_tensor = output_tensor + frame_pos_embed
                    output_tensor, alignment = attention(output_tensor, (keys, values),
                                                         last_attended=last_attended[idx])
                    if self.force_monotonic_attention[idx]:
                        last_attended[idx] = alignment.max(-1)[1].view(-1).data[0]
                    if ave_alignment is None:
                        ave_alignment = alignment
                    else:
                        ave_alignment = ave_alignment + ave_alignment

                # residual
                if isinstance(f, Conv1dGLU):
                    output_tensor = (output_tensor + residual) * math.sqrt(0.5)

            decoder_state = output_tensor
            output_tensor = self.last_conv.incremental_forward(output_tensor, )
            ave_alignment = ave_alignment.div_(num_attention_layers)

            # Ooutput & done flag predictions
            output = F.sigmoid(output_tensor)
            done = F.sigmoid(self.fc(output_tensor))

            decoder_states += [decoder_state]
            outputs += [output]
            alignments += [ave_alignment]
            dones += [done]

            t += 1
            if test_inputs is None:
                if (done > 0.5).all() and t > self.min_decoder_steps:
                    break

        # Remove 1-element time axis
        alignments = list(map(lambda x: x.squeeze(1), alignments))
        decoder_states = list(map(lambda x: x.squeeze(1), decoder_states))
        outputs = list(map(lambda x: x.squeeze(1), outputs))

        # Combine outputs for all time steps
        alignments = torch.stack(alignments).transpose(0, 1)
        decoder_states = torch.stack(decoder_states).transpose(0, 1).contiguous()
        outputs = torch.stack(outputs).transpose(0, 1).contiguous()

        return outputs, alignments, dones, decoder_states

    def start_fresh_sequence(self):
        _clear_modules(self.preattention)
        _clear_modules(self.convolutions)
        self.last_conv.clear_buffer()


def _clear_modules(modules):
    for m in modules:
        try:
            m.clear_buffer()
        except AttributeError:
            pass


class Converter(nn.Module):
    def __init__(self, n_speakers, speaker_embed_dim, in_dim, out_dim, convolutions=((256, 5, 1),) * 4, dropout=0.1):
        super(Converter, self).__init__()
        self.dropout = dropout
        self.in_dim = in_dim
        self.out_dim = out_dim
        self.n_speakers = n_speakers

        # Non causual convolution blocks
        in_channels = convolutions[0][0]
        # Idea from nyanko
        self.convolutions = nn.ModuleList([
            conv1d(in_dim, in_channels, kernel_size=1, padding=0, dilation=1,
                   std_mul=1.0),
            conv_transpose1d(in_channels, in_channels, kernel_size=2,
                             padding=0, stride=2, std_mul=1.0),
            Conv1dGLU(n_speakers, speaker_embed_dim,
                      in_channels, in_channels, kernel_size=3, causal=False,
                      dilation=1, dropout=dropout, std_mul=1.0, residual=True),
            Conv1dGLU(n_speakers, speaker_embed_dim,
                      in_channels, in_channels, kernel_size=3, causal=False,
                      dilation=3, dropout=dropout, std_mul=4.0, residual=True),
            conv_transpose1d(in_channels, in_channels, kernel_size=2,
                             padding=0, stride=2, std_mul=4.0),
            Conv1dGLU(n_speakers, speaker_embed_dim,
                      in_channels, in_channels, kernel_size=3, causal=False,
                      dilation=1, dropout=dropout, std_mul=1.0, residual=True),
            Conv1dGLU(n_speakers, speaker_embed_dim,
                      in_channels, in_channels, kernel_size=3, causal=False,
                      dilation=3, dropout=dropout, std_mul=4.0, residual=True),
        ])

        std_mul = 4.0
        for (out_channels, kernel_size, dilation) in convolutions:
            if in_channels != out_channels:
                self.convolutions.append(
                    conv1d(in_channels, out_channels, kernel_size=1, padding=0,
                           dilation=1, std_mul=std_mul))
                self.convolutions.append(nn.ReLU(inplace=True))
                in_channels = out_channels
                std_mul = 2.0
            self.convolutions.append(
                Conv1dGLU(n_speakers, speaker_embed_dim,
                          in_channels, out_channels, kernel_size, causal=False,
                          dilation=dilation, dropout=dropout, std_mul=std_mul,
                          residual=True))
            in_channels = out_channels
            std_mul = 4.0
        # Last 1x1 convolution
        self.convolutions.append(conv1d(in_channels, out_dim, kernel_size=1,
                                        padding=0, dilation=1, std_mul=std_mul,
                                        dropout=dropout))

    def forward(self, x, speaker_embed=None):
        if self.n_speakers != 1 and speaker_embed == None:
            raise AssertionError("Expected \033[1;31m<self.n_speakers>\033[m to be 1 or \033[1:31m<speaker_embed>[m to be not None, but was not")
        speaker_embed_btc = None
        # Generic case: B x T x C -> B x C x T
        x = x.transpose(1, 2)
        for f in self.convolutions:
            x = f(x, speaker_embed_btc) if isinstance(f, Conv1dGLU) else f(x)
        # Back to B x T x C
        x = x.transpose(1, 2)

        return F.sigmoid(x)



import numpy as np
import os
import tempfile
import torch


from autokeras.constant import Constant
from autokeras.pretrained.base import Pretrained
from autokeras.text.pretrained_bert.tokenization import BertTokenizer
from autokeras.text.pretrained_bert.modeling import BertForSequenceClassification
from autokeras.utils import download_file_from_google_drive, get_device
from torch.utils.data import TensorDataset, DataLoader, SequentialSampler


class InputFeatures(object):

    def __init__(self, input_ids, input_mask, segment_ids):
        self.input_ids = input_ids
        self.input_mask = input_mask
        self.segment_ids = segment_ids


class TextClassifier(Pretrained):

    def __init__(self):
        super(TextClassifier, self).__init__()
        self.device = None
        self.tokenizer = None
        self.model = None
        self.load()

    def load(self):
        self.device = get_device()
        self.tokenizer = BertTokenizer.from_pretrained('bert-base-uncased', do_lower_case=True)

        output_model_file = os.path.join(tempfile.gettempdir(), self.model_dir)

        download_file_from_google_drive(self.file_id, output_model_file)

        model_state_dict = torch.load(output_model_file, map_location=lambda storage, loc: storage)
        self.model = BertForSequenceClassification.from_pretrained('bert-base-uncased', state_dict=model_state_dict, num_labels=self.num_classes)
        self.model.to(self.device)

    def convert_examples_to_features(self, examples, max_seq_length):
        features = []
        for (_, example) in enumerate(examples):
            tokens_a = self.tokenizer.tokenize(example)

            if len(tokens_a) > max_seq_length - 2:
                tokens_a = tokens_a[:(max_seq_length - 2)]

            tokens = ["[CLS]"] + tokens_a + ["[SEP]"]
            segment_ids = [0] * len(tokens)

            input_ids = self.tokenizer.convert_tokens_to_ids(tokens)

            input_mask = [1] * len(input_ids)

            padding = [0] * (max_seq_length - len(input_ids))
            input_ids += padding
            input_mask += padding
            segment_ids += padding

            if len(input_ids) != max_seq_length or len(input_mask) != max_seq_length or len(segment_ids) != max_seq_length:
                raise AssertionError()

            features.append(
                    InputFeatures(input_ids=input_ids,
                                  input_mask=input_mask,
                                  segment_ids=segment_ids))
        return features

    def y_predict(self, x_predict):
        eval_features = self.convert_examples_to_features([x_predict], 128)

        all_input_ids = torch.tensor([f.input_ids for f in eval_features], dtype=torch.long)
        all_input_mask = torch.tensor([f.input_mask for f in eval_features], dtype=torch.long)
        all_segment_ids = torch.tensor([f.segment_ids for f in eval_features], dtype=torch.long)

        eval_data = TensorDataset(all_input_ids, all_input_mask, all_segment_ids)

        eval_sampler = SequentialSampler(eval_data)
        eval_dataloader = DataLoader(eval_data, sampler=eval_sampler, batch_size=1)

        self.model.eval()
        for input_ids, input_mask, segment_ids in eval_dataloader:
            input_ids = input_ids.to(self.device)
            input_mask = input_mask.to(self.device)
            segment_ids = segment_ids.to(self.device)

            with torch.no_grad():
                logits = self.model(input_ids, segment_ids, input_mask)

            logits = logits.detach().cpu().numpy()

            for logit in logits:
                exp = np.exp(logit)
                exp = exp / np.sum(exp)
                y_pred = exp

        return y_pred


class SentimentAnalysis(TextClassifier):

    def __init__(self):
        self.model_dir = 'bert_sentiment_analysis_pytorch_model'
        self.file_id = Constant.SENTIMENT_ANALYSIS_MODEL_ID
        self.num_classes = 2
        super(SentimentAnalysis, self).__init__()

    def predict(self, x_predict):
        y_pred = self.y_predict(x_predict)
        return round(y_pred[1], 2)


class TopicClassifier(TextClassifier):

    def __init__(self):
        self.model_dir = 'bert_topic_classifier_pytorch_model'
        self.file_id = Constant.TOPIC_CLASSIFIER_MODEL_ID
        self.num_classes = 4
        super(TopicClassifier, self).__init__()

    def predict(self, x_predict):
        y_pred = self.y_predict(x_predict)
        class_id = np.argmax(y_pred)
        if class_id == 0:
            return "Business"
        elif class_id == 1:
            return "Sci/Tech"
        elif class_id == 2:
            return "World"
        else:
            return "Sports"




from autokeras.pretrained import VoiceGenerator
from tests.common import TEST_TEMP_DIR, clean_dir
import os


def test_voice_generator():
    voice_generator = VoiceGenerator()
    clean_dir(TEST_TEMP_DIR)
    texts = [
        "Generative adversarial network or variational auto-encoder.",
        "The tuition of the coming semster is 6300 dollars.",
        "The tuition of the coming semster is 6350 dollars.",
        "Turn left on {HH AW1 S S T AH0 N} Street.",
        "This is expensive, it costs me $300.2",
        "This is expensive, it costs me $300",
        "This is cheap, it only costs me $.2",
        "Today he won the 1st prize of the competition",
        "The approximation of pi is 3.14",
    ]

    for idx, text in enumerate(texts):
        save_name = "test_" + str(idx) + ".wav"
        save_name = os.path.join(TEST_TEMP_DIR, save_name)
        voice_generator.generate(text, path=save_name)
    clean_dir(TEST_TEMP_DIR)



from abc import ABC, abstractmethod


class Pretrained(ABC):
    """The base class for all pretrained task."""

    def __init__(self):
        """Initialize the instance."""
        self.model = None

    @abstractmethod
    def load(self, **kwargs):
        """Load pretrained model into self.model."""
        pass

    @abstractmethod
    def predict(self, input_data, **kwargs):
        """Return predict results for the given image
        Returns:
            A numpy.ndarray containing the results.
        """
        pass





from autokeras.pretrained.object_detector import ObjectDetector
from autokeras.pretrained.face_detector import FaceDetector
from autokeras.pretrained.voice_generator.voice_generator import VoiceGenerator


import logging
import os
import queue
import re
import sys
import time
import torch
import torch.multiprocessing as mp


from abc import ABC, abstractmethod
from datetime import datetime
from autokeras.bayesian import BayesianOptimizer
from autokeras.constant import Constant
from autokeras.nn.model_trainer import ModelTrainer
from autokeras.utils import pickle_to_file, pickle_from_file, verbose_print, get_system


class Searcher(ABC):
    """The base class to search for neural architectures.

    This class generate new architectures, call the trainer to train it, and update the optimizer.

    Attributes:
        n_classes: Number of classes in the target classification task.
        input_shape: Arbitrary, although all dimensions in the input shaped must be fixed.
            Use the keyword argument `input_shape` (tuple of integers, does not include the batch axis)
            when using this layer as the first layer in a model.
        verbose: Verbosity mode.
        history: A list that stores the performance of model. Each element in it is a dictionary of 'model_id',
            'loss', and 'metric_value'.
        neighbour_history: A list that stores the performance of neighbor of the best model.
            Each element in it is a dictionary of 'model_id', 'loss', and 'metric_value'.
        path: A string. The path to the directory for saving the searcher.
        metric: An instance of the Metric subclasses.
        loss: A function taking two parameters, the predictions and the ground truth.
        generators: A list of generators used to initialize the search.
        model_count: An integer. the total number of neural networks in the current searcher.
        descriptors: A dictionary of all the neural network architectures searched.
        trainer_args: A dictionary. The params for the constructor of ModelTrainer.
        default_model_len: An integer. Number of convolutional layers in the initial architecture.
        default_model_width: An integer. The number of filters in each layer in the initial architecture.
        training_queue: A list of the generated architectures to be trained.
        x_queue: A list of trained architectures not updated to the gpr.
        y_queue: A list of trained architecture performances not updated to the gpr.
    """

    def __init__(self, n_output_node, input_shape, path, metric, loss, generators, verbose,
                 trainer_args=None,
                 default_model_len=None,
                 default_model_width=None):
        """Initialize the Searcher.

        Args:
            n_output_node: An integer, the number of classes.
            input_shape: A tuple. e.g. (28, 28, 1).
            path: A string. The path to the directory to save the searcher.
            metric: An instance of the Metric subclasses.
            loss: A function taking two parameters, the predictions and the ground truth.
            generators: A list of generators used to initialize the search.
            verbose: A boolean. Whether to output the intermediate information to stdout.
            trainer_args: A dictionary. The params for the constructor of ModelTrainer.
            default_model_len: An integer. Number of convolutional layers in the initial architecture.
            default_model_width: An integer. The number of filters in each layer in the initial architecture.
        """
        if trainer_args is None:
            trainer_args = {}
        self.n_classes = n_output_node
        self.input_shape = input_shape
        self.verbose = verbose
        self.history = []
        self.neighbour_history = []
        self.path = path
        self.metric = metric
        self.loss = loss
        self.generators = generators
        self.model_count = 0
        self.descriptors = []
        self.trainer_args = trainer_args
        self.default_model_len = default_model_len if default_model_len is not None else Constant.MODEL_LEN
        self.default_model_width = default_model_width if default_model_width is not None else Constant.MODEL_WIDTH
        if 'max_iter_num' not in self.trainer_args:
            self.trainer_args['max_iter_num'] = Constant.SEARCH_MAX_ITER

        self.training_queue = []
        self.x_queue = []
        self.y_queue = []

        logging.basicConfig(filename=os.path.join(self.path, datetime.now().strftime('run_%d_%m_%Y : _%H_%M.log')),
                            format='%(asctime)s - %(filename)s - %(message)s', level=logging.DEBUG)

        self._timeout = None


    def load_model_by_id(self, model_id):
        return pickle_from_file(os.path.join(self.path, str(model_id) + '.graph'))

    def load_best_model(self):
        return self.load_model_by_id(self.get_best_model_id())

    def get_metric_value_by_id(self, model_id):
        for item in self.history:
            if item['model_id'] == model_id:
                return item['metric_value']
        return None

    def get_best_model_id(self):
        if self.metric.higher_better():
            return max(self.history, key=lambda x: x['metric_value'])['model_id']
        return min(self.history, key=lambda x: x['metric_value'])['model_id']

    def replace_model(self, graph, model_id):
        pickle_to_file(graph, os.path.join(self.path, str(model_id) + '.graph'))

    def init_search(self):
        """Call the generators to generate the initial architectures for the search."""
        if self.verbose:
            print('\nInitializing search.')
        for generator in self.generators:
            graph = generator(self.n_classes, self.input_shape). \
                generate(self.default_model_len, self.default_model_width)
            model_id = self.model_count
            self.model_count += 1
            self.training_queue.append((graph, -1, model_id))
            self.descriptors.append(graph.extract_descriptor())

        if self.verbose:
            print('Initialization finished.')

    def search(self, train_data, test_data, timeout=60 * 60 * 24):
        """Run the search loop of training, generating and updating once.

        The function will run the training and generate in parallel.
        Then it will update the controller.
        The training is just pop out a graph from the training_queue and train it.
        The generate will call the self.generate function.
        The update will call the self.update function.

        Args:
            train_data: An instance of DataLoader.
            test_data: An instance of Dataloader.
            timeout: An integer, time limit in seconds.
        """
        torch.cuda.empty_cache()
        if not self.history:
            self.init_search()

        self._timeout = time.time() + timeout if timeout is not None else sys.maxsize
        self.trainer_args['timeout'] = timeout
        # Start the new process for training.
        graph, other_info, model_id = self.training_queue.pop(0)
        if self.verbose:
            print('\n')
            print('+' + '-' * 46 + '+')
            print('|' + 'Training model {}'.format(model_id).center(46) + '|')
            print('+' + '-' * 46 + '+')
        # Temporary solution to support GOOGLE Colab
        if get_system() == Constant.SYS_GOOGLE_COLAB:
            # When using Google Colab, use single process for searching and training.
            self.sp_search(graph, other_info, model_id, train_data, test_data)
        else:
            # Use two processes
            self.mp_search(graph, other_info, model_id, train_data, test_data)

    def mp_search(self, graph, other_info, model_id, train_data, test_data):
        ctx = mp.get_context()
        q = ctx.Queue()
        p = ctx.Process(target=train, args=(q, graph, train_data, test_data, self.trainer_args,
                                            self.metric, self.loss, self.verbose, self.path))
        try:
            p.start()
            search_results = self._search_common(q)
            metric_value, loss, graph = q.get(block=True)
            if time.time() >= self._timeout:
                raise TimeoutError
            if self.verbose and search_results:
                for (generated_graph, generated_other_info, new_model_id) in search_results:
                    verbose_print(generated_other_info, generated_graph, new_model_id)

            if metric_value is not None:
                self.add_model(metric_value, loss, graph, model_id)
                self.update(other_info, model_id, graph, metric_value)

        except (TimeoutError, queue.Empty) as e:
            raise TimeoutError from e
        finally:
            # terminate and join the subprocess to prevent any resource leak
            p.terminate()
            p.join()

    def sp_search(self, graph, other_info, model_id, train_data, test_data):
        try:
            metric_value, loss, graph = train(None, graph, train_data, test_data, self.trainer_args,
                                              self.metric, self.loss, self.verbose, self.path)
            # Do the search in current thread.
            search_results = self._search_common()
            if self.verbose and search_results:
                for (generated_graph, generated_other_info, new_model_id) in search_results:
                    verbose_print(generated_other_info, generated_graph, new_model_id)

            if metric_value is not None:
                self.add_model(metric_value, loss, graph, model_id)
                self.update(other_info, model_id, graph, metric_value)

        except TimeoutError as e:
            raise TimeoutError from e

    def _search_common(self, mp_queue=None):
        search_results = []
        if not self.training_queue:
            results = self.generate(mp_queue)
            for (generated_graph, generated_other_info) in results:
                new_model_id = self.model_count
                self.model_count += 1
                self.training_queue.append((generated_graph, generated_other_info, new_model_id))
                self.descriptors.append(generated_graph.extract_descriptor())
                search_results.append((generated_graph, generated_other_info, new_model_id))
            self.neighbour_history = []

        return search_results

    @abstractmethod
    def generate(self, multiprocessing_queue):
        """Generate the next neural architecture.

        Args:
            multiprocessing_queue: the Queue for multiprocessing return value.

        Returns:
            list of 2-element tuples: generated_graph and other_info,
            generated_graph: An instance of Graph.
            other_info: Anything to be saved in the training queue together
            with the architecture.
        """
        pass

    def update(self, *args):
        pass

    def add_model(self, metric_value, loss, graph, model_id):
        """Append the information of evaluated architecture to history."""
        if self.verbose:
            print('\nSaving model.')

        graph.clear_operation_history()
        pickle_to_file(graph, os.path.join(self.path, str(model_id) + '.graph'))

        ret = {'model_id': model_id, 'loss': loss, 'metric_value': metric_value}
        self.neighbour_history.append(ret)
        self.history.append(ret)

        # Update best_model text file
        if model_id == self.get_best_model_id():
            file = open(os.path.join(self.path, 'best_model.txt'), 'w')
            file.write('best model: ' + str(model_id))
            file.close()

        if self.verbose:
            idx = ['model_id', 'loss', 'metric_value']
            header = ['Model ID', 'Loss', 'Metric Value']
            line = '|'.join(x.center(24) for x in header)
            print('+' + '-' * len(line) + '+')
            print('|' + line + '|')

            if self.history:
                r = self.history[-1]
                print('+' + '-' * len(line) + '+')
                line = '|'.join(str(r[x]).center(24) for x in idx)
                print('|' + line + '|')
            print('+' + '-' * len(line) + '+')

        return ret


class BayesianSearcher(Searcher):
    """ Class to search for neural architectures using Bayesian search strategy.

    Attribute:
        optimizer: An instance of BayesianOptimizer.
        t_min: A float. The minimum temperature during simulated annealing.
    """

    def __init__(self, n_output_node, input_shape, path, metric, loss,
                 generators, verbose, trainer_args=None,
                 default_model_len=None, default_model_width=None,
                 t_min=None):
        super(BayesianSearcher, self).__init__(n_output_node, input_shape,
                                               path, metric, loss,
                                               generators, verbose,
                                               trainer_args,
                                               default_model_len,
                                               default_model_width)
        if t_min is None:
            t_min = Constant.T_MIN
        self.optimizer = BayesianOptimizer(self, t_min, metric)

    def generate(self, multiprocessing_queue):
        """Generate the next neural architecture.

        Args:
            multiprocessing_queue: the Queue for multiprocessing return value.

        Returns:
            list of 2-element tuples: generated_graph and other_info,
            for bayesian searcher the length of list is 1.
            generated_graph: An instance of Graph.
            other_info: Anything to be saved in the training queue together with the architecture.

        """
        remaining_time = self._timeout - time.time()
        generated_graph, new_father_id = self.optimizer.generate(self.descriptors,
                                                                 remaining_time, multiprocessing_queue)
        if new_father_id is None:
            new_father_id = 0
            generated_graph = self.generators[0](self.n_classes, self.input_shape). \
                generate(self.default_model_len, self.default_model_width)

        return [(generated_graph, new_father_id)]

    def update(self, other_info, model_id, graph, metric_value):
        """ Update the controller with evaluation result of a neural architecture.

        Args:
            other_info: Anything. In our case it is the father ID in the search tree.
            model_id: An integer.
            graph: An instance of Graph. The trained neural architecture.
            metric_value: The final evaluated metric value.
        """
        father_id = other_info
        self.optimizer.fit([graph.extract_descriptor()], [metric_value])
        self.optimizer.add_child(father_id, model_id)


def train(q, graph, train_data, test_data, trainer_args, metric, loss, verbose, path):
    """Train the neural architecture."""
    try:
        model = graph.produce_model()
        loss, metric_value = ModelTrainer(model=model,
                                          path=path,
                                          train_data=train_data,
                                          test_data=test_data,
                                          metric=metric,
                                          loss_function=loss,
                                          verbose=verbose).train_model(**trainer_args)
        model.set_weight_to_graph()
        if q:
            q.put((metric_value, loss, model.graph))
        return metric_value, loss, model.graph
    except RuntimeError as e:
        if not re.search('out of memory', str(e)):
            raise e
        if verbose:
            print('\nCurrent model size is too big. Discontinuing training this model to search for other models.')
        Constant.MAX_MODEL_SIZE = graph.size() - 1
        if q:
            q.put((None, None, None))
        return None, None, None
    except TimeoutError as exp:
        logging.warning("TimeoutError occurred at train() : {0}".format(str(exp)))
        if q:
            q.put((None, None, None))
        return None, None, None
    except Exception as exp:
        logging.warning("Exception occurred at train() : {0}".format(str(exp)))
        if verbose:
            print("Exception occurred at train() : {0}".format(str(exp)))
        if q:
            q.put((None, None, None))
        return None, None, None




from autokeras.pretrained.voice_generator.deepvoice3_pytorch.model import MultiSpeakerTTSModel, AttentionSeq2Seq


def deepvoice3(n_vocab, embed_dim=256, mel_dim=80, linear_dim=513, r=4, n_speakers=1, speaker_embed_dim=16,
               padding_idx=0, dropout=(1 - 0.95), kernel_size=5, encoder_channels=128, decoder_channels=256,
               converter_channels=256, query_position_rate=1.0, key_position_rate=1.29, use_memory_mask=False,
               trainable_positional_encodings=False, force_monotonic_attention=True,
               use_decoder_state_for_postnet_input=True, max_positions=512, embedding_weight_std=0.1,
               freeze_embedding=False, window_ahead=3, window_backward=1):
    """Build deepvoice3
    """
    from autokeras.pretrained.voice_generator.deepvoice3_pytorch.deepvoice3 import Encoder, Decoder, Converter

    # Seq2seq
    h = encoder_channels  # hidden dim (channels)
    k = kernel_size  # kernel size
    encoder = Encoder(n_vocab, embed_dim, n_speakers=n_speakers, speaker_embed_dim=speaker_embed_dim,
                      padding_idx=padding_idx, embedding_weight_std=embedding_weight_std,
                      convolutions=[(h, k, 1), (h, k, 3), (h, k, 9), (h, k, 27),
                                    (h, k, 1), (h, k, 3), (h, k, 9), (h, k, 27),
                                    (h, k, 1), (h, k, 3)], dropout=dropout)

    h = decoder_channels
    decoder = Decoder(embed_dim, n_speakers=n_speakers, speaker_embed_dim=speaker_embed_dim, in_dim=mel_dim, r=r,
                      max_positions=max_positions, preattention=[(h, k, 1), (h, k, 3)],
                      convolutions=[(h, k, 1), (h, k, 3), (h, k, 9), (h, k, 27),
                                    (h, k, 1)], attention=[True, False, False, False, True], dropout=dropout,
                      use_memory_mask=use_memory_mask, force_monotonic_attention=force_monotonic_attention,
                      query_position_rate=query_position_rate, key_position_rate=key_position_rate,
                      window_ahead=window_ahead, window_backward=window_backward)

    seq2seq = AttentionSeq2Seq(encoder, decoder)

    # Post net
    in_dim = h // r

    h = converter_channels
    converter = Converter(n_speakers=n_speakers, speaker_embed_dim=speaker_embed_dim, in_dim=in_dim, out_dim=linear_dim,
                          convolutions=[(h, k, 1), (h, k, 3), (2 * h, k, 1), (2 * h, k, 3)], dropout=dropout)

    # Seq2seq + post net
    model = MultiSpeakerTTSModel(seq2seq, converter, mel_dim=mel_dim, linear_dim=linear_dim, n_speakers=n_speakers,
                                 speaker_embed_dim=speaker_embed_dim,
                                 trainable_positional_encodings=trainable_positional_encodings,
                                 use_decoder_state_for_postnet_input=use_decoder_state_for_postnet_input,
                                 freeze_embedding=freeze_embedding)

    return model



import numpy as np
from pandas import DataFrame
from scipy.stats import pearsonr

LEVEL_HIGH = 32


def parallel_function(labels, first_batch_keys, task):
    if task == 'label':
        if min(labels) > first_batch_keys:
            labels = labels - np.min(labels)
        return labels.reshape(labels.shape[0], 1)

    elif task == 'frequency':
        cat_dict = {}
        n_rows = labels.shape[0]
        labels = np.expand_dims(labels, axis=1)

        if min(labels) > first_batch_keys:
            labels = labels - np.min(labels)

        frequencies = np.zeros((n_rows, 1))

        for row_index in range(n_rows):
            key = labels[row_index, 0]
            if key in cat_dict:
                cat_dict[key] += 1
            else:
                cat_dict[key] = 1

        n_level = len(cat_dict)
        key_to_frequency = {}

        for key in cat_dict.keys():
            key_to_frequency[key] = cat_dict[key] / n_rows * n_level

        for row_index in range(n_rows):
            key = labels[row_index, 0]
            frequencies[row_index][0] = key_to_frequency[key]

        return frequencies
    elif task == 'num_cat':
        df = DataFrame(data=labels)
        return df.join(df.groupby(1)[0].mean(),
                       rsuffix='r',
                       on=1).values[:, -1:]
    elif task == 'cat_cat':
        df = DataFrame(data=labels)
        df[3] = list(range(len(labels)))
        return df.join(df.groupby([0, 1]).count(),
                       rsuffix='r',
                       on=(0, 1)).values[:, -1:]
    elif task == 'train_num_cat':
        y = first_batch_keys[0]
        df = DataFrame(data=labels)
        fe = df.join(df.groupby(1)[0].mean(),
                     rsuffix='r',
                     on=1).values[:, -1:]
        mu = abs(pearsonr(np.squeeze(np.array(fe)), y)[0])
        if np.isnan(mu):
            mu = 0
        return [[first_batch_keys[1], first_batch_keys[2], mu, mu], first_batch_keys[3]]

    elif task == 'train_cat_cat':
        y = first_batch_keys[0]
        df = DataFrame(data=labels)
        df[3] = list(range(len(labels)))
        fe = df.join(df.groupby([0, 1]).count(),
                     rsuffix='r',
                     on=(0, 1)).values[:, -1:]
        mu = abs(pearsonr(np.squeeze(np.array(fe)), y)[0])
        if np.isnan(mu):
            mu = 0
        return [[first_batch_keys[1], first_batch_keys[2], mu], first_batch_keys[3]]
    return None


def call_parallel(tasks):
    results = []
    for t in tasks:
        results.append(parallel_function(t[0], t[1], t[2]))
    return results


class TabularPreprocessor:
    def __init__(self):
        """
        Initialization function for tabular preprocessor.
        """
        self.num_cat_pair = {}

        self.total_samples = 0

        self.cat_to_int_label = {}
        self.n_first_batch_keys = {}
        self.high_level_cat_keys = []

        self.feature_add_high_cat = 0
        self.feature_add_cat_num = 0
        self.feature_add_cat_cat = 0
        self.order_num_cat_pair = {}

        self.rest = None
        self.budget = None
        self.data_info = None
        self.n_time = None
        self.n_num = None
        self.n_cat = None

    def remove_useless(self, x):
        self.rest = np.where(np.max(x, 0) - np.min(x, 0) != 0)[0]
        return x[:, self.rest]

    def process_time(self, x):
        cols = range(self.n_time)
        if len(cols) > 10:
            cols = cols[:10]
        x_time = x[:, cols]
        for i in cols:
            for j in range(i + 1, len(cols)):
                x = np.append(x, np.expand_dims(x_time[:, i] - x_time[:, j], 1), 1)
        return x

    def extract_data(self, raw_x):
        # only get numerical variables
        ret = np.concatenate([raw_x['TIME'], raw_x['NUM'], raw_x['CAT']], axis=1)
        n_rows = ret.shape[0]
        n_num_col = ret.shape[1] - self.n_cat

        n_cat_col = self.n_cat
        if n_cat_col <= 0:
            return ret.astype(np.float64)

        # preprocess (multi-value) categorical data
        for col_index in range(n_num_col, n_num_col + n_cat_col):
            for row_index in range(n_rows):
                key = str(ret[row_index, col_index])
                if key in self.cat_to_int_label[col_index]:
                    ret[row_index, col_index] = self.cat_to_int_label[col_index][key]
                    continue
                new_value = len(self.cat_to_int_label[col_index])
                self.cat_to_int_label[col_index][key] = new_value
                ret[row_index, col_index] = new_value

        return ret.astype(np.float64)

    def cat_to_num(self, x, y=None):
        if y is not None:
            mark = self.n_time + self.n_num

            for col_index in range(self.n_time + self.n_num, self.n_time + self.n_num + self.n_cat):
                if self.n_first_batch_keys[col_index] <= LEVEL_HIGH:
                    self.num_cat_pair[mark] = (col_index,)
                    mark += 1
                else:
                    self.num_cat_pair[mark] = (col_index, col_index)
                    mark += 1

            mark_1 = 0
            tasks = []
            for i, cat_col_index1 in enumerate(self.high_level_cat_keys):
                for cat_col_index2 in self.high_level_cat_keys[i + 1:]:
                    tasks.append((x[:, (cat_col_index1, cat_col_index2)],
                                  [y, cat_col_index1, cat_col_index2, mark_1],
                                  'train_cat_cat'))
                    mark_1 += 1

            all_results = call_parallel(tasks)

            num_cat_pair_1 = {}
            pearsonr_dict_1 = {}
            for result in all_results:
                if result[0][-1] > 0.001:
                    pearsonr_dict_1[result[1]] = result[0][-1]
                    num_cat_pair_1[result[1]] = result[0]
            pearsonr_high_1 = sorted(pearsonr_dict_1, key=pearsonr_dict_1.get, reverse=True)[:self.feature_add_cat_cat]
            num_cat_pair_1 = {key: num_cat_pair_1[key] for key in pearsonr_high_1}
            num_cat_pair_1 = {i + mark: num_cat_pair_1[key] for i, key in enumerate(num_cat_pair_1)}
            self.num_cat_pair.update(num_cat_pair_1)
            mark += len(pearsonr_high_1)

            mark_2 = 0
            tasks_2 = []
            for cat_col_index in self.high_level_cat_keys:
                for num_col_index in range(self.n_time, self.n_time + self.n_num):
                    tasks_2.append((x[:, (num_col_index, cat_col_index)],
                                    [y, num_col_index, cat_col_index, mark_2],
                                    'train_num_cat'))
                    mark_2 += 1

            all_results = call_parallel(tasks_2)

            num_cat_pair_2 = {}
            pearsonr_dict_2 = {}
            for result in all_results:
                if result[0][-1] > 0.001:
                    pearsonr_dict_2[result[1]] = result[0][-1]
                    num_cat_pair_2[result[1]] = result[0]
            pearsonr_high_2 = sorted(pearsonr_dict_2, key=pearsonr_dict_2.get, reverse=True)[:self.feature_add_cat_num]
            num_cat_pair_2 = {key: num_cat_pair_2[key] for key in pearsonr_high_2}
            num_cat_pair_2 = {i + mark: num_cat_pair_2[key] for i, key in enumerate(num_cat_pair_2)}
            self.num_cat_pair.update(num_cat_pair_2)
            self.order_num_cat_pair = sorted(list(self.num_cat_pair.keys()))
            print('num_cat_pair_2:', num_cat_pair_2)

        tasks = []
        for key in self.order_num_cat_pair:
            if len(self.num_cat_pair[key]) == 1:
                (col_index,) = self.num_cat_pair[key]
                tasks.append((x[:, col_index], self.n_first_batch_keys[col_index], 'label'))
            if len(self.num_cat_pair[key]) == 2:
                (col_index, col_index) = self.num_cat_pair[key]
                tasks.append((x[:, col_index], self.n_first_batch_keys[col_index], 'frequency'))
            if len(self.num_cat_pair[key]) == 3:
                (cat_col_index1, cat_col_index2, mu) = self.num_cat_pair[key]
                tasks.append((x[:, (cat_col_index1,
                                    cat_col_index2)], self.n_first_batch_keys[cat_col_index1], 'cat_cat'))
            elif len(self.num_cat_pair[key]) == 4:
                (num_col_index, cat_col_index, mu, a) = self.num_cat_pair[key]
                tasks.append((x[:, (num_col_index, cat_col_index)], self.n_first_batch_keys[cat_col_index], 'num_cat'))

        results = call_parallel(tasks)
        all_num = x.shape[1] - self.n_cat
        results = [x[:, :all_num]] + results
        ret = np.concatenate(results, axis=1)

        return ret

    def fit(self, raw_x, y, time_limit, data_info):
        """
        This function should train the model parameters.

        Args:
            raw_x: a numpy.ndarray instance containing the training data.
            y: training label vector.
            time_limit: remaining time budget.
            data_info: meta-features of the dataset, which is an numpy.ndarray describing the
             feature type of each column in raw_x. The feature type include:
                     'TIME' for temporal feature, 'NUM' for other numerical feature,
                     and 'CAT' for categorical feature.
        """
        # Get Meta-Feature
        self.budget = time_limit
        self.data_info = data_info if data_info is not None else self.extract_data_info(raw_x)
        print('QQ: {}'.format(self.data_info))

        self.n_time = sum(self.data_info == 'TIME')
        self.n_num = sum(self.data_info == 'NUM')
        self.n_cat = sum(self.data_info == 'CAT')

        self.total_samples = raw_x.shape[0]

        print('QQ1: {}'.format(self.n_time))
        print('QQ2: {}'.format(self.n_num))
        print('QQ3: {}'.format(self.n_cat))
        raw_x = {'TIME': raw_x[:, self.data_info == 'TIME'],
                 'NUM': raw_x[:, self.data_info == 'NUM'],
                 'CAT': raw_x[:, self.data_info == 'CAT']}


        for col_index in range(self.n_num + self.n_time, self.n_num + self.n_time + self.n_cat):
            self.cat_to_int_label[col_index] = {}

        x = self.extract_data(raw_x)

        d_size = x.shape[0] * x.shape[1] / self.budget
        if d_size > 35000:
            self.feature_add_high_cat = 0
        else:
            self.feature_add_high_cat = 10

        # Iterate cat features
        for col_index in range(self.n_num + self.n_time, self.n_num + self.n_time + self.n_cat):
            self.n_first_batch_keys[col_index] = len(self.cat_to_int_label[col_index])
        high_level_cat_keys_tmp = sorted(self.n_first_batch_keys, key=self.n_first_batch_keys.get, reverse=True)[
                                  :self.feature_add_high_cat]
        for i in high_level_cat_keys_tmp:
            if self.n_first_batch_keys[i] > 1e2:
                self.high_level_cat_keys.append(i)

        # Convert NaN to zeros
        x = np.nan_to_num(x)

        # Encode high-order categorical data to numerical with frequency
        x = self.cat_to_num(x, y)

        x = self.process_time(x)
        x = self.remove_useless(x)

        return x

    def encode(self, raw_x, time_limit=None):
        """
        This function should train the model parameters.

        Args:
            raw_x: a numpy.ndarray instance containing the training/testing data.
            time_limit: remaining time budget.
        Both inputs X and y are numpy arrays.
        If fit is called multiple times on incremental data (train, test1, test2, etc.)
        you should warm-start your training from the pre-trained model. Past data will
        NOT be available for re-training.
        """
        # Get Meta-Feature
        if time_limit is None:
            if self.budget is None:
                time_limit = 24 * 60 * 60
                self.budget = time_limit
        else:
            self.budget = time_limit

        raw_x = {'TIME': raw_x[:, self.data_info == 'TIME'],
                 'NUM': raw_x[:, self.data_info == 'NUM'],
                 'CAT': raw_x[:, self.data_info == 'CAT']}
        x = self.extract_data(raw_x)

        # Convert NaN to zeros
        x = np.nan_to_num(x)

        # Encode high-order categorical data to numerical with frequency
        x = self.cat_to_num(x)

        x = self.process_time(x)
        if self.rest is not None:
            x = x[:, self.rest]
        return x

    @staticmethod
    def extract_data_info(raw_x):
        """
        This function extracts the data info automatically based on the type of each feature in raw_x.

        Args:
            raw_x: a numpy.ndarray instance containing the training data.
        """
        data_info = []
        row_num, col_num = raw_x.shape
        for col_idx in range(col_num):
            try:
                raw_x[:, col_idx].astype(np.float)
                data_info.append('NUM')
            except:
                data_info.append('CAT')
        return np.array(data_info)



# coding: utf-8

import torch
from torch import nn
import math
import numpy as np
from torch.nn import functional as F


def position_encoding_init(n_position, d_pos_vec, position_rate=1.0):
    """Init the sinusoid position encoding table """

    # keep dim 0 for padding token position encoding zero vector
    position_enc = np.array([
        [position_rate * pos / np.power(10000, 2 * (i // 2) / d_pos_vec) for i in range(d_pos_vec)]
        if pos != 0 else np.zeros(d_pos_vec) for pos in range(n_position)])

    position_enc = torch.from_numpy(position_enc).float()

    return position_enc


def sinusoidal_encode(x, w):
    y = w * x
    y[1:, 0::2] = torch.sin(y[1:, 0::2].clone())
    y[1:, 1::2] = torch.cos(y[1:, 1::2].clone())
    return y


class SinusoidalEncoding(nn.Embedding):

    def __init__(self, num_embeddings, embedding_dim,
                 *args, **kwargs):
        super(SinusoidalEncoding, self).__init__(num_embeddings, embedding_dim,
                                                 padding_idx=0,
                                                 *args, **kwargs)
        self.weight.data = position_encoding_init(num_embeddings, embedding_dim, position_rate=1.0)

    def forward(self, x, w=1.0):
        isscaler = np.isscalar(w)
        assert self.padding_idx is not None

        if isscaler or w.size(0) == 1:
            weight = sinusoidal_encode(self.weight, w)
            return F.embedding(
                x, weight, self.padding_idx, self.max_norm,
                self.norm_type, self.scale_grad_by_freq, self.sparse)


def linear(in_features, out_features, dropout=0):
    """Weight-normalized Linear layer (input: N x T x C)"""
    m = nn.Linear(in_features, out_features)
    m.weight.data.normal_(mean=0, std=math.sqrt((1 - dropout) / in_features))
    m.bias.data.zero_()
    return nn.utils.weight_norm(m)


def embedding(num_embeddings, embedding_dim, padding_idx, std=0.01):
    m = nn.Embedding(num_embeddings, embedding_dim, padding_idx=padding_idx)
    m.weight.data.normal_(0, std)
    return m


def conv1d(in_channels, out_channels, kernel_size, dropout=0, std_mul=4.0, **kwargs):
    from .conv import Conv1d
    m = Conv1d(in_channels, out_channels, kernel_size, **kwargs)
    std = math.sqrt((std_mul * (1.0 - dropout)) / (m.kernel_size[0] * in_channels))
    m.weight.data.normal_(mean=0, std=std)
    m.bias.data.zero_()
    return nn.utils.weight_norm(m)


def conv_transpose1d(in_channels, out_channels, kernel_size, dropout=0,
                     std_mul=1.0, **kwargs):
    m = nn.ConvTranspose1d(in_channels, out_channels, kernel_size, **kwargs)
    std = math.sqrt((std_mul * (1.0 - dropout)) / (m.kernel_size[0] * in_channels))
    m.weight.data.normal_(mean=0, std=std)
    m.bias.data.zero_()
    return nn.utils.weight_norm(m)


class Conv1dGLU(nn.Module):
    """(Dilated) Conv1d + Gated linear unit + (optionally) speaker embedding
    """

    def __init__(self, n_speakers, speaker_embed_dim,
                 in_channels, out_channels, kernel_size,
                 dropout, padding=None, dilation=1, causal=False, residual=False,
                 *args, **kwargs):
        super(Conv1dGLU, self).__init__()
        self.dropout = dropout
        self.residual = residual
        if padding is None:
            # no future time stamps available
            if causal:
                padding = (kernel_size - 1) * dilation
            else:
                padding = (kernel_size - 1) // 2 * dilation
        self.causal = causal

        self.conv = conv1d(in_channels, 2 * out_channels, kernel_size,
                           dropout=dropout, padding=padding, dilation=dilation,
                           *args, **kwargs)
        if n_speakers > 1:
            self.speaker_proj = linear(speaker_embed_dim, out_channels)
        else:
            self.speaker_proj = None

    def forward(self, x, speaker_embed=None):
        return self._forward(x, False)

    def incremental_forward(self, x):
        return self._forward(x, True)

    def _forward(self, x, is_incremental):
        residual = x
        x = F.dropout(x, p=self.dropout, training=self.training)
        if is_incremental:
            splitdim = -1
            x = self.conv.incremental_forward(x, )
        else:
            splitdim = 1
            x = self.conv(x)
            # remove future time steps
            x = x[:, :, :residual.size(-1)] if self.causal else x

        a, b = x.split(x.size(splitdim) // 2, dim=splitdim)
        x = a * F.sigmoid(b)
        return (x + residual) * math.sqrt(0.5) if self.residual else x

    def clear_buffer(self):
        self.conv.clear_buffer()



from autokeras.pretrained.object_detector import ObjectDetector
from tests.common import TEST_TEMP_DIR, clean_dir


def test_object_detection():
    detector = ObjectDetector()
    detector.load()
    img_path = 'tests/resources/images_test/od.JPG'
    result = detector.predict(img_path, TEST_TEMP_DIR)
    assert isinstance(result, list)
    clean_dir(TEST_TEMP_DIR)



import os
import re

import numpy as np
import tensorflow as tf
from keras import Input, Model
from keras import backend
from keras.layers import Embedding

from autokeras.constant import Constant
from autokeras.utils import download_file_with_extract, temp_path_generator, ensure_dir


def download_pre_train(file_path, extract_path):
    """Download pre train file from link in constant.py.

    Args:
        file_path: String, contains download file path + file name.
        extract_path: String extract path name.
    """
    file_link = Constant.PRE_TRAIN_FILE_LINK
    print("try downloading pre train weights from link %s" % file_link)
    download_file_with_extract(file_link, file_path=file_path, extract_path=extract_path)


def clean_str(string):
    """Tokenization/string cleaning for all string.

    Original taken from https://github.com/yoonkim/CNN_sentence/blob/master/process_data.py
    """
    string = re.sub(r"[^A-Za-z0-9(),!?\'\`]", " ", string)
    string = re.sub(r"\'s", " \'s", string)
    string = re.sub(r"\'ve", " \'ve", string)
    string = re.sub(r"n\'t", " n\'t", string)
    string = re.sub(r"\'re", " \'re", string)
    string = re.sub(r"\'d", " \'d", string)
    string = re.sub(r"\'ll", " \'ll", string)
    string = re.sub(r",", " , ", string)
    string = re.sub(r"!", " ! ", string)
    string = re.sub(r"\(", " \( ", string)
    string = re.sub(r"\)", " \) ", string)
    string = re.sub(r"\?", " \? ", string)
    string = re.sub(r"\s{2,}", " ", string)
    return string.strip().lower()


def tokenlize_text(max_num_words, max_seq_length, x_train):
    """Tokenlize text.

    Vectorize a text corpus by transform each text in texts to a sequence of integers.

    Args:
        max_num_words: Int, max number of words in the dictionary.
        max_seq_length: Int, the length of each text sequence, padding if shorter, trim is longer.
        x_train: List contains text data.

    Returns:
        x_train: Tokenlized input data.
        word_index: Dictionary contains word with tokenlized index.
    """
    from keras_preprocessing.sequence import pad_sequences
    from keras_preprocessing.text import Tokenizer
    print("tokenlizing texts...")
    tokenizer = Tokenizer(num_words=max_num_words)
    tokenizer.fit_on_texts(x_train)
    sequences = tokenizer.texts_to_sequences(x_train)
    word_index = tokenizer.word_index
    x_train = pad_sequences(sequences, maxlen=max_seq_length)
    print("data readed and convert to %d length sequences" % max_seq_length)
    return x_train, word_index


def read_embedding_index(extract_path):
    """Read pre train file convert to embedding vector.

    Read the pre trained file into a dictionary where key is the word and value is embedding vector.

    Args:
        extract_path: String contains pre trained file path.

    Returns:
        embedding_index: Dictionary contains word with pre trained index.
    """
    embedding_index = {}
    f = open(os.path.join(extract_path, Constant.PRE_TRAIN_FILE_NAME), encoding="utf-8")
    for line in f:
        values = line.split()
        word = values[0]
        coefs = np.asarray(values[1:], dtype='float32')
        embedding_index[word] = coefs
    f.close()
    return embedding_index


def load_pretrain(path, word_index):
    """Load the pretrain file into embedding weights.

    This method will first generate the embedding index and then generate
    embedding matrix according to the word_index.

    Args:
        path: String, path to store the pretrain files.
        word_index: Dictionary contains word with tokenlized index.

    Returns:
        embedding_matrix: Numpy array as the pretrain model embedding layer weights.
    """
    print("loading pretrain weights...")
    file_path = os.path.join(path, Constant.FILE_PATH)
    extract_path = os.path.join(path, Constant.EXTRACT_PATH)
    download_pre_train(file_path=file_path, extract_path=extract_path)
    embedding_index = read_embedding_index(extract_path)
    print('Total %s word vectors embedded.' % len(embedding_index))

    # convert the pretrained embedding index to weights
    embedding_matrix = np.random.random((len(word_index) + 1, Constant.EMBEDDING_DIM))
    for word, i in word_index.items():
        embedding_vector = embedding_index.get(word)
        if embedding_vector is not None:
            embedding_matrix[i] = embedding_vector
    return embedding_matrix


def processing(path, word_index, input_length, x_train):
    """Processing string array with pretrained vectors.

    convert an n dimension string array into n * k * m dimension float numpy array. Each k * m array represents
    a string. k is the input_length which means an upper bound of the string length, for string shorter than
    k will be pad and longer string will be cropped. m is defined by the pretrained file.

    Args:
        path: String, path where the pre trained files stored.
        word_index: Dictionary, contains word with tokenlized index.
        input_length: Int, an upper bound of the string length.
        x_train: String array.

    Returns:
        x_train: Numpy array as processed x_train.
    """

    embedding_matrix = load_pretrain(path=path, word_index=word_index)

    config = tf.ConfigProto(allow_soft_placement=True)
    config.gpu_options.allow_growth = True
    sess = tf.Session(config=config)
    backend.set_session(sess)
    print("generating preprocessing model...")
    embedding_layer = Embedding(len(word_index) + 1,
                                Constant.EMBEDDING_DIM,
                                weights=[embedding_matrix],
                                input_length=input_length,
                                trainable=False)

    sequence_input = Input(shape=(input_length,), dtype='int32')
    embedded_sequences = embedding_layer(sequence_input)
    model = Model(sequence_input, embedded_sequences)
    print("converting text to vector...")
    x_train = model.predict(x_train)
    del model

    return x_train


def text_preprocess(x_train):
    """This is the text preprocess main method.

    It takes an raw string, clean it and processing it into tokenlized numpy array.
    """
    if Constant.STORE_PATH == '':
        temp_path = temp_path_generator()
        path = temp_path + '_store'
    else:
        path = Constant.STORE_PATH

    ensure_dir(path)

    x_train = [clean_str(x) for x in x_train]
    x_train, word_index = tokenlize_text(max_seq_length=Constant.MAX_SEQUENCE_LENGTH,
                                         max_num_words=Constant.MAX_NB_WORDS,
                                         x_train=x_train)

    print("generating preprocessing model...")
    x_train = processing(path=path, word_index=word_index, input_length=Constant.MAX_SEQUENCE_LENGTH, x_train=x_train)
    return x_train



class Constant:
    # Data

    VALIDATION_SET_SIZE = 0.08333
    CUTOUT_HOLES = 1
    CUTOUT_RATIO = 0.5

    # Searcher

    MAX_MODEL_NUM = 1000
    BETA = 2.576
    KERNEL_LAMBDA = 1.0
    T_MIN = 0.0001
    N_NEIGHBOURS = 8
    MAX_MODEL_SIZE = (1 << 25)
    MAX_LAYER_WIDTH = 4096
    MAX_LAYERS = 200

    # Grid Dimensions
    LENGTH_DIM = 0
    WIDTH_DIM = 1

    # Default Search Space
    DEFAULT_LENGTH_SEARCH = [50, 75, 100]
    DEFAULT_WIDTH_SEARCH = [64, 128, 256]

    # Model Defaults

    DENSE_DROPOUT_RATE = 0.5
    CONV_DROPOUT_RATE = 0.25
    MLP_DROPOUT_RATE = 0.25
    CONV_BLOCK_DISTANCE = 2
    DENSE_BLOCK_DISTANCE = 1
    MODEL_LEN = 3
    MLP_MODEL_LEN = 3
    MLP_MODEL_WIDTH = 5
    MODEL_WIDTH = 64
    POOLING_KERNEL_SIZE = 2

    # ModelTrainer

    DATA_AUGMENTATION = True
    MAX_ITER_NUM = 200
    MIN_LOSS_DEC = 1e-4
    MAX_NO_IMPROVEMENT_NUM = 5
    MAX_BATCH_SIZE = 128
    LIMIT_MEMORY = False
    SEARCH_MAX_ITER = 200

    # text preprocessor

    EMBEDDING_DIM = 100
    MAX_SEQUENCE_LENGTH = 400
    MAX_NB_WORDS = 5000
    EXTRACT_PATH = "glove/"
    STORE_PATH = ''

    # Download file name

    FILE_PATH = "glove.zip"
    PRE_TRAIN_FILE_LINK = "http://nlp.stanford.edu/data/glove.6B.zip"
    PRE_TRAIN_FILE_NAME = "glove.6B.100d.txt"

    PRE_TRAIN_DETECTION_FILE_LINK = "https://s3.amazonaws.com/amdegroot-models/ssd300_mAP_77.43_v2.pth"

    PRE_TRAIN_VOICE_GENERATOR_MODEL_GOOGLE_DRIVE_ID = "1E-B92LZz4dgg8DU81D6pyhOzM9yvvBTj"
    PRE_TRAIN_VOICE_GENERATOR_MODEL_NAME = "20180505_deepvoice3_checkpoint_step000640000.pth"
    PRE_TRAIN_VOICE_GENERATOR_SAVE_FILE_DEFAULT_NAME = "test.wav"

    # constants for pretrained model of face detection
    FACE_DETECTOR = {
        'MODEL_GOOGLE_ID': [
            '1QJWKpAHRrAjrYPl6hQNDaoyBjoa_LRgz',
            '10aCiR393E6TLkp9KPPl4JhZamYqUVBO1',
            '1RRBtPlzw46peS-A8pyYGsPRHHFIUrSVV'
        ],
        'MODEL_NAMES': [
            'pnet.pt',
            'rnet.pt',
            'onet.pt'
        ]
    }

    OBJECT_DETECTOR = {
        'MODEL_GOOGLE_ID': '1QGG1trfj-z5_2OGNoSarUB4wx81cG-sa',
        'MODEL_NAME': 'object_detection_pretrained.pth'
    }

    # Constants for pretrained models of Sentiment Analysis and Topic Classification.

    SENTIMENT_ANALYSIS_MODEL_ID = '1flRlQjfIa2toQ6HNmInhqrh4NuxGh8pT'
    TOPIC_CLASSIFIER_MODEL_ID = '1U7C3xPid1ZvBKpkfW9KikrmNui0yJqnk'

    # Image Resize

    MAX_IMAGE_SIZE = 128 * 128

    # SYS Constant

    SYS_LINUX = 'linux'
    SYS_WINDOWS = 'windows'
    SYS_GOOGLE_COLAB = 'goog_colab'

    # Google drive downloader
    CHUNK_SIZE = 32768
    DOWNLOAD_URL = "https://docs.google.com/uc?export=download"



from autokeras.constant import Constant
from autokeras.search import Searcher
from autokeras.utils import assert_search_space


class GridSearcher(Searcher):
    """ Class to search for neural architectures using Greedy search strategy.

    Attribute:
        search_space: A dictionary. Specifies the search dimensions and their possible values
    """
    def __init__(self, n_output_node, input_shape, path, metric, loss, generators, verbose, search_space={},
                 trainer_args=None, default_model_len=None, default_model_width=None):
        super(GridSearcher, self).__init__(n_output_node, input_shape, path, metric, loss, generators, verbose,
                                           trainer_args, default_model_len, default_model_width)
        self.search_space, self.search_dimensions = assert_search_space(search_space)
        self.search_space_counter = 0

    def get_search_dimensions(self):
        return self.search_dimensions

    def search_space_exhausted(self):
        """ Check if Grid search has exhausted the search space """
        if self.search_space_counter == len(self.search_dimensions):
            return True
        return False

    def search(self, train_data, test_data, timeout=60 * 60 * 24):
        """Run the search loop of training, generating and updating once.

        Call the base class implementation for search with

        Args:
            train_data: An instance of DataLoader.
            test_data: An instance of Dataloader.
            timeout: An integer, time limit in seconds.
        """
        if self.search_space_exhausted():
            return
        else:
            super().search(train_data, test_data, timeout)

    def update(self, other_info, model_id, graph, metric_value):
        return

    def generate(self, multiprocessing_queue):
        """Generate the next neural architecture.

        Args:
            multiprocessing_queue: the Queue for multiprocessing return value.

        Returns:
            list of 2-element tuples: generated_graph and other_info,
            for grid searcher the length of list is 1.
            generated_graph: An instance of Graph.
            other_info: Always 0.
        """
        grid = self.get_grid()
        self.search_space_counter += 1
        generated_graph = self.generators[0](self.n_classes, self.input_shape). \
            generate(grid[Constant.LENGTH_DIM], grid[Constant.WIDTH_DIM])
        return [(generated_graph, 0)]

    def get_grid(self):
        """ Return the next grid to be searched """
        if self.search_space_counter < len(self.search_dimensions):
            return self.search_dimensions[self.search_space_counter]
        return None


import pytest
import numpy as np

from autokeras.tabular.tabular_supervised import *

from tests.common import clean_dir, TEST_TEMP_DIR


def test_fit_evalute_predict_binary_classification():
    train_x = None
    train_y = None
    clean_dir(TEST_TEMP_DIR)
    clf = TabularClassifier(path=TEST_TEMP_DIR)
    nsample = 59
    feat_exist_ind = [0, 5]
    for ntime in feat_exist_ind:
        for nnum in feat_exist_ind:
            for ncat in feat_exist_ind:
                    datainfo = np.array(['TIME'] * ntime + ['NUM'] * nnum + ['CAT'] * ncat)
                    num_feat = ntime + nnum + ncat
                    x_num = np.random.random([nsample, nnum])
                    x_time = np.random.random([nsample, ntime])
                    x_cat = np.random.randint(0, 40, [nsample, ncat])

                    train_x = np.concatenate([x_num, x_time, x_cat], axis=1)
                    train_y = np.random.randint(0, 2, nsample)
                    print(datainfo)
                    if num_feat == 0:
                        with pytest.raises(ValueError):
                            clf.fit(train_x, train_y, data_info=datainfo)
                    else:
                        clf.fit(train_x, train_y, data_info=datainfo)
                        results = clf.predict(train_x)
                        assert all(map(lambda result: result in train_y, results))
    score = clf.evaluate(train_x, train_y)
    assert score >= 0.0
    clean_dir(TEST_TEMP_DIR)


def test_fit_evalute_predict_multiclass_classification():
    clean_dir(TEST_TEMP_DIR)
    clf = TabularClassifier(path=TEST_TEMP_DIR)
    clf.verbose = True
    nsample = 1000
    [ntime, nnum, ncat] = [11, 15, 13]
    datainfo = np.array(['TIME'] * ntime + ['NUM'] * nnum + ['CAT'] * ncat)
    x_num = np.random.random([nsample, nnum])
    x_time = np.random.random([nsample, ntime])
    x_cat = np.random.randint(0, 200, [nsample, ncat])
    train_x = np.concatenate([x_num, x_time, x_cat], axis=1)
    train_y = np.random.randint(0, 3, nsample)
    clf.fit(train_x, train_y, data_info=datainfo)
    results = clf.predict(train_x)
    assert all(map(lambda result: result in train_y, results))
    score = clf.evaluate(train_x, train_y)
    assert score >= 0.0
    clean_dir(TEST_TEMP_DIR)


def test_fit_predict_evalute_regression():
    clean_dir(TEST_TEMP_DIR)
    clf = TabularRegressor(path=TEST_TEMP_DIR)
    nsample = 1000
    [ntime, nnum, ncat] = [3, 15, 3]
    datainfo = np.array(['TIME'] * ntime + ['NUM'] * nnum + ['CAT'] * ncat)
    x_num = np.random.random([nsample, nnum])
    x_time = np.random.random([nsample, ntime])
    x_cat = np.random.randint(0, 200, [nsample, ncat])
    train_x = np.concatenate([x_num, x_time, x_cat], axis=1)
    train_y = train_x[:, 5]
    clf.fit(train_x, train_y, data_info=datainfo)
    results = clf.predict(train_x)
    assert len(results) == len(train_x)
    score = clf.evaluate(train_x, train_y)
    assert score >= 0.0
    clean_dir(TEST_TEMP_DIR)



import time
from copy import deepcopy
from multiprocessing.queues import Queue

from autokeras.bayesian import contain, SearchTree
from autokeras.net_transformer import transform
from autokeras.search import Searcher


class GreedyOptimizer:

    def __init__(self, searcher, metric):
        self.searcher = searcher
        self.metric = metric
        self.search_tree = SearchTree()

    def generate(self, descriptors, timeout, sync_message):
        """Generate new neighbor architectures from the best model.

        Args:
            descriptors: All the searched neural architectures.
            timeout: An integer. The time limit in seconds.
            sync_message: the Queue for multiprocessing return value.

        Returns:
            out: A list of 2-elements tuple. Each tuple contains
                an instance of Graph, a morphed neural network with weights
                and the father node id in the search tree.
        """
        out = []
        start_time = time.time()
        descriptors = deepcopy(descriptors)

        if isinstance(sync_message, Queue) and sync_message.qsize() != 0:
            return out
        model_id = self.searcher.get_neighbour_best_model_id()
        graph = self.searcher.load_model_by_id(model_id)
        father_id = model_id
        for temp_graph in transform(graph):
            if contain(descriptors, temp_graph.extract_descriptor()):
                continue
            out.append((deepcopy(temp_graph), father_id))
        remaining_time = timeout - (time.time() - start_time)

        if remaining_time < 0:
            raise TimeoutError
        return out

    def add_child(self, father_id, model_id):
        self.search_tree.add_child(father_id, model_id)


class GreedySearcher(Searcher):
    """ Class to search for neural architectures using Greedy search strategy.

    Attribute:
        optimizer: An instance of BayesianOptimizer.
    """

    def __init__(self, n_output_node, input_shape, path, metric, loss, generators, verbose,
                 trainer_args=None,
                 default_model_len=None,
                 default_model_width=None):
        super(GreedySearcher, self).__init__(n_output_node, input_shape,
                                             path, metric, loss, generators,
                                             verbose, trainer_args, default_model_len,
                                             default_model_width)
        self.optimizer = GreedyOptimizer(self, metric)

    def generate(self, multiprocessing_queue):
        """Generate the next neural architecture.

        Args:
            multiprocessing_queue: the Queue for multiprocessing return value.
                pass into the search algorithm for synchronizing

        Returns:
            results: A list of 2-element tuples. Each tuple contains an instance of Graph,
                and anything to be saved in the training queue together with the architecture

        """
        remaining_time = self._timeout - time.time()
        results = self.optimizer.generate(self.descriptors, remaining_time,
                                          multiprocessing_queue)
        if not results:
            new_father_id = 0
            generated_graph = self.generators[0](self.n_classes, self.input_shape). \
                generate(self.default_model_len, self.default_model_width)
            results.append((generated_graph, new_father_id))

        return results

    def update(self, other_info, model_id, *args):
        """ Update the controller with evaluation result of a neural architecture.

        Args:
            other_info: Anything. In our case it is the father ID in the search tree.
            model_id: An integer.
        """
        father_id = other_info
        self.optimizer.add_child(father_id, model_id)

    def load_neighbour_best_model(self):
        return self.load_model_by_id(self.get_neighbour_best_model_id())

    def get_neighbour_best_model_id(self):
        if self.metric.higher_better():
            return max(self.neighbour_history, key=lambda x: x['metric_value'])['model_id']
        return min(self.neighbour_history, key=lambda x: x['metric_value'])['model_id']


# coding: utf-8
from torch import nn
from torch.nn import functional as F


class Conv1d(nn.Conv1d):
    """Extended nn.Conv1d for incremental dilated convolutions
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.clear_buffer()
        self._linearized_weight = None

    def incremental_forward(self, input_data):

        # reshape weight
        weight = self._get_linearized_weight()
        kw = self.kernel_size[0]
        dilation = self.dilation[0]

        bsz = input_data.size(0)  # conv_input: bsz x len x dim
        if kw > 1:
            input_data = input_data.data
            if self.input_buffer is None:
                self.input_buffer = input_data.new(bsz, kw + (kw - 1) * (dilation - 1), input_data.size(2))
                self.input_buffer.zero_()
            else:
                # shift buffer
                self.input_buffer[:, :-1, :] = self.input_buffer[:, 1:, :].clone()
            # append next input
            self.input_buffer[:, -1, :] = input_data[:, -1, :]
            input_data = self.input_buffer
            if dilation > 1:
                input_data = input_data[:, 0::dilation, :].contiguous()
        input_data = F.linear(input_data.view(bsz, -1), weight, self.bias)
        return input_data.view(bsz, 1, -1)

    def clear_buffer(self):
        self.input_buffer = None

    def _get_linearized_weight(self):
        if self._linearized_weight is None:
            kw = self.kernel_size[0]
            # nn.Conv1d
            weight = self.weight.transpose(1, 2).contiguous()

            assert weight.size() == (self.out_channels, kw, self.in_channels)
            self._linearized_weight = weight.view(self.out_channels, -1)
        return self._linearized_weight



from abc import abstractmethod

import os
from lightgbm import LGBMClassifier, LGBMRegressor
from sklearn.model_selection import RandomizedSearchCV
from sklearn.model_selection import StratifiedKFold, KFold
from sklearn.metrics import roc_auc_score, f1_score, mean_squared_error
import numpy as np
import random

from autokeras.supervised import Supervised
from autokeras.tabular.tabular_preprocessor import TabularPreprocessor
from autokeras.utils import rand_temp_folder_generator, ensure_dir


class TabularSupervised(Supervised):
    def __init__(self, path=None, **kwargs):
        """
        Initialization function for tabular supervised learner.
        """
        super().__init__(**kwargs)
        self.is_trained = False
        self.clf = None
        self.objective = None
        self.tabular_preprocessor = None
        self.path = path if path is not None else rand_temp_folder_generator()
        ensure_dir(self.path)
        if self.verbose:
            print('Path:', path)
        self.save_filename = os.path.join(self.path, 'lgbm.txt')
        self.time_limit = None
        self.lgbm = None

    def search(self, search_space, search_iter, n_estimators, x, y):
        if 'n_estimators' in search_space:
            del search_space['n_estimators']
        params = {
            'boosting_type': ['gbdt'],
            'min_child_weight': [5],
            'min_split_gain': [1.0],
            'subsample': [0.8],
            'colsample_bytree': [0.6],
            'max_depth': [10],
            'n_estimators': n_estimators,
            'num_leaves': [70],
            'learning_rate': [0.04],
        }
        params.update(search_space)
        if self.verbose:
            print(params)
        folds = 3
        score_metric, skf = self.get_skf(folds)

        random_search = RandomizedSearchCV(self.lgbm, param_distributions=params, n_iter=search_iter,
                                           scoring=score_metric,
                                           n_jobs=1, cv=skf, verbose=0, random_state=1001)

        random_search.fit(x, y)
        self.clf = random_search.best_estimator_

        return random_search.best_params_

    @abstractmethod
    def get_skf(self, folds):
        pass

    def fit(self, x, y, time_limit=None, data_info=None):
        """
        This function should train the model parameters.

        Args:
            x: A numpy.ndarray instance containing the training data.
            y: training label vector.
            time_limit: remaining time budget.
            data_info: meta-features of the dataset, which is an numpy.ndarray describing the
             feature type of each column in raw_x. The feature type include:
                     'TIME' for temporal feature, 'NUM' for other numerical feature,
                     and 'CAT' for categorical feature.
        Both inputs X and y are numpy arrays.
        If fit is called multiple times on incremental data (train, test1, test2, etc.)
        you should warm-start your training from the pre-trained model. Past data will
        NOT be available for re-training.
        """

        if time_limit is None:
            time_limit = 24 * 60 * 60
        self.time_limit = time_limit

        self.init_lgbm(y)

        self.tabular_preprocessor = TabularPreprocessor()

        if x.shape[1] == 0:
            raise ValueError("No feature exist!")

        x = self.tabular_preprocessor.fit(x, y, self.time_limit, data_info)

        if x.shape[0] > 600:
            grid_train_percentage = max(600.0 / x.shape[0], 0.1)
        else:
            grid_train_percentage = 1
        grid_n = int(x.shape[0] * grid_train_percentage)
        idx = random.sample(list(range(x.shape[0])), grid_n)

        grid_train_x = x[idx, :]
        grid_train_y = y[idx]

        while x.shape[0] < 60:
            x = np.concatenate([x, x], axis=0)
            y = np.concatenate([y, y], axis=0)

        response_rate = sum(y) / len(y)

        if not self.is_trained:
            # Two-step cross-validation for hyperparameter selection
            if self.verbose:
                print('-----------------Search Regularization Params---------------------')
            if response_rate < 0.005:
                depth_choice = [5]
            else:
                depth_choice = [8, 10]

            params = {
                'min_split_gain': [0.1],
                'max_depth': depth_choice,
                'min_child_weight': [5, 10, 30, 50, 60, 80, 100],
                'colsample_bytree': [0.6, 0.7],
                'learning_rate': [0.3],
                'subsample': [0.8],
                'num_leaves': [80],
            }

            search_iter = 14
            n_estimators_choice = [50]
            best_param = self.search(
                params,
                search_iter,
                n_estimators_choice,
                grid_train_x, grid_train_y)

            if self.verbose:
                print('-----------------Search Learning Rate---------------------')
            for key, value in best_param.items():
                best_param[key] = [value]
            best_param['learning_rate'] = [0.03, 0.045, 0.06, 0.075, 0.85, 0.95, 0.105, 0.12]
            n_estimators_choice = [100, 150, 200]
            search_iter = 16

            self.search(
                best_param,
                search_iter,
                n_estimators_choice,
                grid_train_x, grid_train_y)

            if self.verbose:
                print('self.clf', self.clf)
            self.is_trained = True

        # Fit Model
        self.clf.fit(x, y)

        self.clf.booster_.save_model(self.save_filename)

        if self.verbose:
            print("The whole available data is: ")
            print("Real-FIT: dim(X)= [{:d}, {:d}]".format(x.shape[0], x.shape[1]))

            print('Feature Importance:')
            print(self.clf.feature_importances_)

    @abstractmethod
    def init_lgbm(self, y):
        pass
    def predict(self, x_test):
        """
        This function should provide predictions of labels on (test) data.
        The function predict eventually casdn return probabilities or continuous values.
        """
        x_test = self.tabular_preprocessor.encode(x_test)
        y = self.clf.predict(x_test)
        if y is None:
            raise ValueError("Tabular predictor does not exist")
        return y

    @abstractmethod
    def evaluate(self, x_test, y_test):
        pass

    def final_fit(self, x_train, y_train, x_test=None, y_test=None, trainer_args=None, retrain=False):
        x_train = self.tabular_preprocessor.encode(x_train)
        self.clf.fit(x_train, y_train)


class TabularRegressor(TabularSupervised):
    """TabularRegressor class.
    It is used for tabular data regression with lightgbm regressor.
    """
    def __init__(self, path=None):
        super().__init__(path)
        self.objective = 'regression'

    def evaluate(self, x_test, y_test):
        y_pred = self.predict(x_test)
        return mean_squared_error(y_test, y_pred)

    def init_lgbm(self, y):
        self.lgbm = LGBMRegressor(silent=False,
                                  verbose=-1,
                                  n_jobs=1,
                                  objective=self.objective)

    def get_skf(self, folds):
        return 'neg_mean_squared_error', KFold(n_splits=folds, shuffle=True, random_state=1001)


class TabularClassifier(TabularSupervised):
    """TabularClassifier class.
     It is used for tabular data classification with lightgbm classifier.
    """

    def init_lgbm(self, y):
        n_classes = len(set(y))
        if n_classes == 2:
            self.objective = 'binary'
            self.lgbm = LGBMClassifier(silent=False,
                                       verbose=-1,
                                       n_jobs=1,
                                       objective=self.objective)
        else:
            self.objective = 'multiclass'
            self.lgbm = LGBMClassifier(silent=False,
                                       verbose=-1,
                                       n_jobs=1,
                                       num_class=n_classes,
                                       objective=self.objective)

    def evaluate(self, x_test, y_test):
        if self.verbose:
            print('objective:', self.objective)
        y_pred = self.predict(x_test)
        results = None
        if self.objective == 'binary':
            results = roc_auc_score(y_test, y_pred)
        elif self.objective == 'multiclass':
            results = f1_score(y_test, y_pred, average='weighted')
        return results

    def get_skf(self, folds):
        if self.lgbm.objective == 'binary':
            score_metric = 'roc_auc'
            skf = StratifiedKFold(n_splits=folds, shuffle=True, random_state=1001)
        else:
            score_metric = 'f1_weighted'
            skf = StratifiedKFold(n_splits=folds, shuffle=True, random_state=1001)
        return score_metric, skf


from abc import abstractmethod
from sklearn.model_selection import train_test_split
import numpy as np
from autokeras.utils import validate_xy, resize_image_data, compute_image_resize_params
from autokeras.nn.metric import Accuracy
from autokeras.nn.loss_function import classification_loss
from autokeras.nn.generator import ResNetGenerator, DenseNetGenerator
from autokeras.search import train
from autokeras.constant import Constant
from autokeras.preprocessor import ImageDataTransformer, OneHotEncoder
from autokeras.supervised import SingleModelSupervised


class PredefinedModel(SingleModelSupervised):
    """The base class for the predefined model without architecture search

    Attributes:
        graph: The graph form of the model.
        y_encoder: Label encoder, used in transform_y or inverse_transform_y for encode the label. For example,
                   if one hot encoder needed, y_encoder can be OneHotEncoder.
        data_transformer: A instance of transformer to process the data, See example as ImageDataTransformer.
        verbose: A boolean of whether the search process will be printed to stdout.
        path: A string. The path to a directory, where the intermediate results are saved.
    """

    def __init__(self, y_encoder=OneHotEncoder(), data_transformer=None, verbose=False, path=None):
        super().__init__(verbose, path)
        self.graph = None
        self.generator = None
        self.y_encoder = y_encoder
        self.data_transformer = data_transformer

    @abstractmethod
    def _init_generator(self, n_output_node, input_shape):
        """Initialize the generator to generate the model architecture.

        Args:
            n_output_node:  A integer value represent the number of output node in the final layer.
            input_shape: A tuple to express the shape of every train entry.
        """
        pass

    @property
    def loss(self):
        return classification_loss

    @property
    def metric(self):
        return Accuracy

    def preprocess(self, x):
        return resize_image_data(x, self.resize_shape)

    def transform_y(self, y_train):
        return self.y_encoder.transform(y_train)

    def inverse_transform_y(self, output):
        return self.y_encoder.inverse_transform(output)

    def fit(self, x, y, trainer_args=None):
        """Trains the model on the dataset given.

        Args:
            x: A numpy.ndarray instance containing the training data or the training data combined with the
               validation data.
            y: A numpy.ndarray instance containing the label of the training data. or the label of the training data
               combined with the validation label.
            trainer_args: A dictionary containing the parameters of the ModelTrainer constructor.
        """
        validate_xy(x, y)
        self.resize_shape = compute_image_resize_params(x)
        x = self.preprocess(x)
        self.y_encoder.fit(y)
        y = self.transform_y(y)
        # Divide training data into training and testing data.
        validation_set_size = int(len(y) * Constant.VALIDATION_SET_SIZE)
        validation_set_size = min(validation_set_size, 500)
        validation_set_size = max(validation_set_size, 1)
        x_train, x_test, y_train, y_test = train_test_split(x, y,
                                                            test_size=validation_set_size,
                                                            random_state=42)

        # initialize data_transformer
        self.data_transformer = ImageDataTransformer(x_train)
        # Wrap the data into DataLoaders
        train_loader = self.data_transformer.transform_train(x_train, y_train)
        test_loader = self.data_transformer.transform_test(x_test, y_test)

        self.generator = self._init_generator(self.y_encoder.n_classes, x_train.shape[1:])
        graph = self.generator.generate()

        if trainer_args is None:
            trainer_args = {'max_no_improvement_num': 30}
        _, _1, self.graph = train(None, graph, train_loader, test_loader,
                                  trainer_args, self.metric, self.loss,
                                  self.verbose, self.path)


class PredefinedResnet(PredefinedModel):
    def _init_generator(self, n_output_node, input_shape):
        return ResNetGenerator(n_output_node, input_shape)


class PredefinedDensenet(PredefinedModel):
    def _init_generator(self, n_output_node, input_shape):
        return DenseNetGenerator(n_output_node, input_shape)



from autokeras.pretrained.text_classifier import TopicClassifier

def test_topic_classifier():
    topic_classifier = TopicClassifier()

    topic_name = topic_classifier.predict("Risk mitigation is the pursuit of opportunities where the potential upside is far greater than the potential downside")

    if topic_name != "Business":
        raise AssertionError()

    topic_name = topic_classifier.predict("With a tap on the screen the app will recognise your face and bring up the filter menu")

    if topic_name != "Sci/Tech":
        raise AssertionError()

    topic_name = topic_classifier.predict("Anthony received a loud ovation when he was shown on the overhead videoboard in the first quarter")

    if topic_name != "Sports":
        raise AssertionError()

    topic_name = topic_classifier.predict("The soviet union was created about five years after Russian Revolution.")

    print(topic_name)
    if topic_name != "World":
        raise AssertionError()



import pytest

from autokeras.tabular.tabular_preprocessor import *

from tests.common import clean_dir, TEST_TEMP_DIR


def test_extract_data_info():
    clean_dir(TEST_TEMP_DIR)
    tp = TabularPreprocessor()
    nsample = 10000
    half_sample = 5000
    [tp.n_num, tp.n_cat] = [5, 1]
    x_num = np.random.random([nsample, tp.n_num])
    x_cat = np.array(['a'] * half_sample + ['b'] * half_sample).reshape([nsample, 1])
    raw_x = np.concatenate([x_num, x_cat], axis=1)
    data_info = tp.extract_data_info(raw_x)
    ground_truth = np.array(['NUM'] * tp.n_num + ['CAT'] * tp.n_cat)
    assert (data_info == ground_truth).all()
    clean_dir(TEST_TEMP_DIR)


# ----------------------------------
# The idea of the classes and functions in this file is largely borrowed from
# https://github.com/amdegroot/ssd.pytorch
# A huge thank you to the authors: Max deGroot and Ellis Brown
# modified by: Wuyang Chen, Haifeng Jin
# ----------------------------------

import os
from itertools import product as product
from math import sqrt as sqrt

import cv2
import numpy as np
import torch
from matplotlib import pyplot as plt
from torch import nn as nn
from torch.autograd import Variable, Function
from torch.nn import functional
from torch.nn import init as init

from autokeras.constant import Constant
from autokeras.pretrained.base import Pretrained
from autokeras.utils import get_device, temp_path_generator, ensure_dir, download_file_from_google_drive

"""VOC Dataset Classes

Original author: Francisco Massa
https://github.com/fmassa/vision/blob/voc_dataset/torchvision/datasets/voc.py

Updated by: Ellis Brown, Max deGroot
"""

# gets home dir cross platform
HOME = os.path.expanduser("~")

# for making bounding boxes pretty
COLORS = ((255, 0, 0, 128), (0, 255, 0, 128), (0, 0, 255, 128),
          (0, 255, 255, 128), (255, 0, 255, 128), (255, 255, 0, 128))

MEANS = (104, 117, 123)

# SSD300 CONFIGS
VOC = {
    'num_classes': 21,
    'lr_steps': (80000, 100000, 120000),
    'max_iter': 120000,
    'feature_maps': [38, 19, 10, 5, 3, 1],
    'min_dim': 300,
    'steps': [8, 16, 32, 64, 100, 300],
    'min_sizes': [30, 60, 111, 162, 213, 264],
    'max_sizes': [60, 111, 162, 213, 264, 315],
    'aspect_ratios': [[2], [2, 3], [2, 3], [2, 3], [2], [2]],
    'variance': [0.1, 0.2],
    'clip': True,
    'name': 'VOC',
}

COCO = {
    'num_classes': 91,
    'lr_steps': (280000, 360000, 400000),
    'max_iter': 400000,
    'feature_maps': [38, 19, 10, 5, 3, 1],
    'min_dim': 300,
    'steps': [8, 16, 32, 64, 100, 300],
    'min_sizes': [21, 45, 99, 153, 207, 261],
    'max_sizes': [45, 99, 153, 207, 261, 315],
    'aspect_ratios': [[2], [2, 3], [2, 3], [2, 3], [2], [2]],
    'variance': [0.1, 0.2],
    'clip': True,
    'name': 'COCO',
}

VOC_CLASSES = (  # always index 0
    'aeroplane', 'bicycle', 'bird', 'boat',
    'bottle', 'bus', 'car', 'cat', 'chair',
    'cow', 'diningtable', 'dog', 'horse',
    'motorbike', 'person', 'pottedplant',
    'sheep', 'sofa', 'train', 'tvmonitor')

# note: if you used our download scripts, this should be right
VOC_ROOT = os.path.join(HOME, "object_detection/data/VOCdevkit/")


class SSD(nn.Module):
    """Single Shot Multibox Architecture
    The network is composed of a base VGG network followed by the
    added multibox conv layers.  Each multibox layer branches into
        1) conv2d for class conf scores
        2) conv2d for localization predictions
        3) associated priorbox layer to produce default bounding
           boxes specific to the layer's feature map size.
    See: https://arxiv.org/pdf/1512.02325.pdf for more details.

    Args:
        phase: (string) Can be "test" or "train"
        size: input image size
        base: VGG16 layers for input, size of either 300 or 500
        extras: extra layers that feed to multibox loc and conf layers
        head: "multibox head" consists of loc and conf conv layers
    """

    def __init__(self, phase, size, base, extras, head, num_classes, device):
        super(SSD, self).__init__()
        self.phase = phase
        self.num_classes = num_classes
        self.cfg = (COCO, VOC)[num_classes == 21]
        self.priorbox = PriorBox(self.cfg)
        self.device = device
        self.priors = Variable(self.priorbox.forward(), volatile=True).to(self.device)
        self.size = size

        # SSD network
        self.vgg = nn.ModuleList(base)
        # Layer learns to scale the l2 normalized features from conv4_3
        self.L2Norm = L2Norm(512, 20)
        self.extras = nn.ModuleList(extras)

        self.loc = nn.ModuleList(head[0])
        self.conf = nn.ModuleList(head[1])

        if phase == 'test':
            self.softmax = nn.Softmax(dim=-1)
            self.detect = Detect(num_classes, 0, 200, 0.01, 0.45)

    def forward(self, x):
        """Applies network layers and ops on input image(s) x.

        Args:
            x: input image or batch of images. Shape: [batch,3,300,300].

        Return:
            Depending on phase:
            test:
                Variable(tensor) of output class label predictions,
                confidence score, and corresponding location predictions for
                each object detected. Shape: [batch,topk,7]

            train:
                list of concat outputs from:
                    1: confidence layers, Shape: [batch*num_priors,num_classes]
                    2: localization layers, Shape: [batch,num_priors*4]
                    3: priorbox layers, Shape: [2,num_priors*4]
        """
        sources = list()
        loc = list()
        conf = list()

        # apply vgg up to conv4_3 relu
        for k in range(23):
            x = self.vgg[k](x)

        s = self.L2Norm(x)
        sources.append(s)

        # apply vgg up to fc7
        for k in range(23, len(self.vgg)):
            x = self.vgg[k](x)
        sources.append(x)

        # apply extra layers and cache source layer outputs
        for k, v in enumerate(self.extras):
            x = functional.relu(v(x), inplace=True)
            if k % 2 == 1:
                sources.append(x)

        # apply multibox head to source layers
        for (x, l, c) in zip(sources, self.loc, self.conf):
            loc.append(l(x).permute(0, 2, 3, 1).contiguous())
            conf.append(c(x).permute(0, 2, 3, 1).contiguous())

        loc = torch.cat([o.view(o.size(0), -1) for o in loc], 1)
        conf = torch.cat([o.view(o.size(0), -1) for o in conf], 1)
        if self.phase == "test":
            output = self.detect(
                loc.view(loc.size(0), -1, 4).to(self.device),  # loc preds
                self.softmax(conf.view(conf.size(0), -1,
                                       self.num_classes)).to(self.device),  # conf preds
                self.priors.type(type(x.data)).to(self.device)  # default boxes
            )
        else:
            output = (
                loc.view(loc.size(0), -1, 4),
                conf.view(conf.size(0), -1, self.num_classes),
                self.priors
            )
        return output

    def load_weights(self, base_file):
        other, ext = os.path.splitext(base_file)
        if ext == '.pkl' or '.pth':
            print('Loading weights into state dict...')
            self.load_state_dict(torch.load(base_file,
                                            map_location=lambda storage, loc: storage))
            print('Finished!')
        else:
            print('Sorry only .pth and .pkl files supported.')


# This function is derived from torchvision VGG make_layers()
# https://github.com/pytorch/vision/blob/master/torchvision/models/vgg.py
def vgg(cfg, i, batch_norm=False):
    layers = []
    in_channels = i
    for v in cfg:
        if v == 'M':
            layers += [nn.MaxPool2d(kernel_size=2, stride=2)]
        elif v == 'C':
            layers += [nn.MaxPool2d(kernel_size=2, stride=2, ceil_mode=True)]
        else:
            conv2d = nn.Conv2d(in_channels, v, kernel_size=3, padding=1)
            if batch_norm:
                layers += [conv2d, nn.BatchNorm2d(v), nn.ReLU(inplace=True)]
            else:
                layers += [conv2d, nn.ReLU(inplace=True)]
            in_channels = v
    pool5 = nn.MaxPool2d(kernel_size=3, stride=1, padding=1)
    conv6 = nn.Conv2d(512, 1024, kernel_size=3, padding=6, dilation=6)
    conv7 = nn.Conv2d(1024, 1024, kernel_size=1)
    layers += [pool5, conv6,
               nn.ReLU(inplace=True), conv7, nn.ReLU(inplace=True)]
    return layers


def add_extras(cfg, i):
    # Extra layers added to VGG for feature scaling
    layers = []
    in_channels = i
    flag = False
    for k, v in enumerate(cfg):
        if in_channels != 'S':
            if v == 'S':
                layers += [nn.Conv2d(in_channels, cfg[k + 1],
                                     kernel_size=(1, 3)[flag], stride=2, padding=1)]
            else:
                layers += [nn.Conv2d(in_channels, v, kernel_size=(1, 3)[flag])]
            flag = not flag
        in_channels = v
    return layers


def multi_box(vgg_result, extra_layers, cfg, num_classes):
    loc_layers = []
    conf_layers = []
    vgg_source = [21, -2]
    for k, v in enumerate(vgg_source):
        loc_layers += [nn.Conv2d(vgg_result[v].out_channels,
                                 cfg[k] * 4, kernel_size=3, padding=1)]
        conf_layers += [nn.Conv2d(vgg_result[v].out_channels,
                                  cfg[k] * num_classes, kernel_size=3, padding=1)]
    for k, v in enumerate(extra_layers[1::2], 2):
        loc_layers += [nn.Conv2d(v.out_channels, cfg[k]
                                 * 4, kernel_size=3, padding=1)]
        conf_layers += [nn.Conv2d(v.out_channels, cfg[k]
                                  * num_classes, kernel_size=3, padding=1)]
    return vgg_result, extra_layers, (loc_layers, conf_layers)


class L2Norm(nn.Module):
    def __init__(self, n_channels, scale):
        super(L2Norm, self).__init__()
        self.n_channels = n_channels
        self.gamma = scale or None
        self.eps = 1e-10
        self.weight = nn.Parameter(torch.Tensor(self.n_channels))
        self.reset_parameters()

    def reset_parameters(self):
        init.constant(self.weight, self.gamma)

    def forward(self, x):
        norm = x.pow(2).sum(dim=1, keepdim=True).sqrt() + self.eps
        # x /= norm
        x = torch.div(x, norm)
        out = self.weight.unsqueeze(0).unsqueeze(2).unsqueeze(3).expand_as(x) * x
        return out


class PriorBox(object):
    """Compute priorbox coordinates in center-offset form for each source
    feature map.
    """

    def __init__(self, cfg):
        super(PriorBox, self).__init__()
        self.image_size = cfg['min_dim']
        # number of priors for feature map location (either 4 or 6)
        self.num_priors = len(cfg['aspect_ratios'])
        self.variance = cfg['variance'] or [0.1]
        self.feature_maps = cfg['feature_maps']
        self.min_sizes = cfg['min_sizes']
        self.max_sizes = cfg['max_sizes']
        self.steps = cfg['steps']
        self.aspect_ratios = cfg['aspect_ratios']
        self.clip = cfg['clip']
        self.version = cfg['name']
        for v in self.variance:
            if v <= 0:
                raise ValueError('Variances must be greater than 0')

    def forward(self):
        mean = []
        for k, f in enumerate(self.feature_maps):
            for i, j in product(range(f), repeat=2):
                f_k = self.image_size / self.steps[k]
                # unit center x,y
                cx = (j + 0.5) / f_k
                cy = (i + 0.5) / f_k

                # aspect_ratio: 1
                # rel size: min_size
                s_k = self.min_sizes[k] / self.image_size
                mean += [cx, cy, s_k, s_k]

                # aspect_ratio: 1
                # rel size: sqrt(s_k * s_(k+1))
                s_k_prime = sqrt(s_k * (self.max_sizes[k] / self.image_size))
                mean += [cx, cy, s_k_prime, s_k_prime]

                # rest of aspect ratios
                for ar in self.aspect_ratios[k]:
                    mean += [cx, cy, s_k * sqrt(ar), s_k / sqrt(ar)]
                    mean += [cx, cy, s_k / sqrt(ar), s_k * sqrt(ar)]
        # back to torch land
        output = torch.Tensor(mean).view(-1, 4)
        if self.clip:
            output.clamp_(max=1, min=0)
        return output


class Detect(Function):
    """At test time, Detect is the final layer of SSD.  Decode location preds,
    apply non-maximum suppression to location predictions based on conf
    scores and threshold to a top_k number of output predictions for both
    confidence score and locations.
    """

    @staticmethod
    def backward(ctx, *grad_outputs):
        pass

    def __init__(self, num_classes, bkg_label, top_k, conf_thresh, nms_thresh):
        self.num_classes = num_classes
        self.background_label = bkg_label
        self.top_k = top_k
        # Parameters used in nms.
        self.nms_thresh = nms_thresh
        if nms_thresh <= 0:
            raise ValueError('nms_threshold must be non negative.')
        self.conf_thresh = conf_thresh
        self.variance = VOC['variance']

    def forward(self, loc_data, conf_data, prior_data):
        """
        Args:
            loc_data: (tensor) Loc preds from loc layers
                Shape: [batch,num_priors*4]
            conf_data: (tensor) Shape: Conf preds from conf layers
                Shape: [batch*num_priors,num_classes]
            prior_data: (tensor) Prior boxes and variances from priorbox layers
                Shape: [1,num_priors,4]
        """
        num = loc_data.size(0)  # batch size
        num_priors = prior_data.size(0)
        output = torch.zeros(num, self.num_classes, self.top_k, 5)
        conf_preds = conf_data.view(num, num_priors,
                                    self.num_classes).transpose(2, 1)

        # Decode predictions into bboxes.
        for i in range(num):
            decoded_boxes = decode(loc_data[i], prior_data, self.variance)
            # For each class, perform nms
            conf_scores = conf_preds[i].clone()

            for cl in range(1, self.num_classes):
                c_mask = conf_scores[cl].gt(self.conf_thresh)
                scores = conf_scores[cl][c_mask]
                # if scores.dim() == 0:
                if scores.size(0) == 0:
                    continue
                l_mask = c_mask.unsqueeze(1).expand_as(decoded_boxes)
                boxes = decoded_boxes[l_mask].view(-1, 4)
                # idx of highest scoring and non-overlapping boxes per class
                ids, count = nms(boxes, scores, self.nms_thresh, self.top_k)
                output[i, cl, :count] = \
                    torch.cat((scores[ids[:count]].unsqueeze(1),
                               boxes[ids[:count]]), 1)
        flt = output.contiguous().view(num, -1, 5)
        _, idx = flt[:, :, 0].sort(1, descending=True)
        _, rank = idx.sort(1)
        flt[(rank < self.top_k).unsqueeze(-1).expand_as(flt)].fill_(0)
        return output


# Adapted from https://github.com/Hakuyume/chainer-ssd
def decode(loc, priors, variances):
    """Decode locations from predictions using priors to undo
    the encoding we did for offset regression at train time.
    Args:
        loc (tensor): location predictions for loc layers,
            Shape: [num_priors,4]
        priors (tensor): Prior boxes in center-offset form.
            Shape: [num_priors,4].
        variances: (list[float]) Variances of priorboxes
    Return:
        decoded bounding box predictions
    """

    boxes = torch.cat((
        priors[:, :2] + loc[:, :2] * variances[0] * priors[:, 2:],
        priors[:, 2:] * torch.exp(loc[:, 2:] * variances[1])), 1)
    boxes[:, :2] -= boxes[:, 2:] / 2
    boxes[:, 2:] += boxes[:, :2]
    return boxes


# Original author: Francisco Massa:
# https://github.com/fmassa/object-detection.torch
# Ported to PyTorch by Max deGroot (02/01/2017)
def nms(boxes, scores, overlap=0.5, top_k=200):
    """Apply non-maximum suppression at test time to avoid detecting too many
    overlapping bounding boxes for a given object.
    Args:
        boxes: (tensor) The location preds for the img, Shape: [num_priors,4].
        scores: (tensor) The class predscores for the img, Shape:[num_priors].
        overlap: (float) The overlap thresh for suppressing unnecessary boxes.
        top_k: (int) The Maximum number of box preds to consider.
    Return:
        The indices of the kept boxes with respect to num_priors.
    """

    keep = scores.new(scores.size(0)).zero_().long()
    if boxes.numel() == 0:
        return keep
    x1 = boxes[:, 0]
    y1 = boxes[:, 1]
    x2 = boxes[:, 2]
    y2 = boxes[:, 3]
    area = torch.mul(x2 - x1, y2 - y1)
    v, idx = scores.sort(0)  # sort in ascending order
    # I = I[v >= 0.01]
    idx = idx[-top_k:]  # indices of the top-k largest vals
    xx1 = boxes.new()
    yy1 = boxes.new()
    xx2 = boxes.new()
    yy2 = boxes.new()
    w = boxes.new()
    h = boxes.new()

    # keep = torch.Tensor()
    count = 0
    while idx.numel() > 0:
        i = idx[-1]  # index of current largest val
        # keep.append(i)
        keep[count] = i
        count += 1
        if idx.size(0) == 1:
            break
        idx = idx[:-1]  # remove kept element from view
        # load bboxes of next highest vals
        torch.index_select(x1, 0, idx, out=xx1)
        torch.index_select(y1, 0, idx, out=yy1)
        torch.index_select(x2, 0, idx, out=xx2)
        torch.index_select(y2, 0, idx, out=yy2)
        # store element-wise max with next highest score
        xx1 = torch.clamp(xx1, min=x1[i])
        yy1 = torch.clamp(yy1, min=y1[i])
        xx2 = torch.clamp(xx2, max=x2[i])
        yy2 = torch.clamp(yy2, max=y2[i])
        w.resize_as_(xx2)
        h.resize_as_(yy2)
        w = xx2 - xx1
        h = yy2 - yy1
        # check sizes of xx1 and xx2.. after each iteration
        w = torch.clamp(w, min=0.0)
        h = torch.clamp(h, min=0.0)
        inter = w * h
        # iou = i / (area(a) + area(b) - i)
        rem_areas = torch.index_select(area, 0, idx)  # load remaining areas)
        union = (rem_areas - inter) + area[i]
        iou = inter / union  # store result in iou
        # keep only elements with an iou <= overlap
        idx = idx[iou.le(overlap)]
    return keep, count


class ObjectDetector(Pretrained):
    def __init__(self):
        super(ObjectDetector, self).__init__()
        self.model = None
        self.device = get_device()

    def _build_ssd(self, phase, size=300, num_classes=21):
        if phase != "test" and phase != "train":
            print("ERROR: Phase: " + phase + " not recognized")
            return
        if size != 300:
            print("ERROR: You specified size " + repr(size) + ". However, " +
                  "currently only SSD300 (size=300) is supported!")
            return
        base = {
            '300': [64, 64, 'M', 128, 128, 'M', 256, 256, 256, 'C', 512, 512, 512, 'M',
                    512, 512, 512],
            '512': [],
        }
        extras = {
            '300': [256, 'S', 512, 128, 'S', 256, 128, 256, 128, 256],
            '512': [],
        }
        mbox = {
            '300': [4, 6, 6, 6, 4, 4],  # number of boxes per feature map location
            '512': [],
        }

        base_, extras_, head_ = multi_box(vgg(base[str(size)], 3),
                                          add_extras(extras[str(size)], 1024),
                                          mbox[str(size)], num_classes)
        return SSD(phase, size, base_, extras_, head_, num_classes, self.device)

    def load(self, model_path=None):

        if model_path is None:
            model_file_name = Constant.OBJECT_DETECTOR['MODEL_NAME']
            temp_path = temp_path_generator()
            ensure_dir(temp_path)
            model_path = f'{temp_path}/{model_file_name}'
            download_file_from_google_drive(file_id=Constant.OBJECT_DETECTOR['MODEL_GOOGLE_ID'], dest_path=model_path)
        # load net
        num_classes = len(VOC_CLASSES) + 1  # +1 for background
        self.model = self._build_ssd('test', 300, num_classes)  # initialize SSD
        if self.device.startswith("cuda"):
            self.model.load_state_dict(torch.load(model_path))
        else:
            self.model.load_state_dict(torch.load(model_path, map_location=lambda storage, loc: storage))
        self.model.eval()
        print('Finished loading model!')

        self.model = self.model.to(self.device)

    def predict(self, input_data, output_file_path=None):
        """
        
        Returns:
            List of dictionaries. Each dictionary is like
            {"left": int, "top": int, "width": int, "height": int: "category": str, "confidence": float}
        """
        from matplotlib.ticker import NullLocator

        dataset_mean = (104, 117, 123)

        image = cv2.imread(input_data, cv2.IMREAD_COLOR)
        height, width, _ = image.shape
        rgb_image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        x = base_transform(rgb_image, 300, dataset_mean)
        x = x.astype(np.float32)
        x = torch.from_numpy(x).permute(2, 0, 1)
        xx = Variable(x.unsqueeze(0))  # wrap tensor in Variable
        # if self.device.startswith("cuda"):
        xx = xx.to(self.device)
        y = self.model(xx)

        # (batch, num_classes, top_k, 5), 5 means (confidence, )
        detections = y.data
        results = []
        # scale each detection back up to the image
        scale = torch.Tensor(rgb_image.shape[1::-1]).repeat(2)
        for i in range(detections.size(1)):
            j = 0
            while detections[0, i, j, 0] >= 0.6:
                score = detections[0, i, j, 0].item()
                label_name = VOC_CLASSES[i - 1]
                pt = (detections[0, i, j, 1:] * scale).cpu().numpy()
                # result = ((pt[0], pt[1]), (pt[2] - pt[0] + 1, pt[3] - pt[1] + 1), label_name, score)
                result = {
                          "left": max(int(np.round(pt[0])), 0),
                          "top": max(int(np.round(pt[1])), 0),
                          "width": min(int(np.round(pt[2] - pt[0] + 1)), width),
                          "height": min(int(np.round(pt[3] - pt[1] + 1)), height),
                          "category": label_name,
                          "confidence": score
                         }
                results.append(result)
                j += 1

        if output_file_path is not None:
            # plt.figure(figsize=(10,10))
            colors = plt.cm.hsv(np.linspace(0, 1, 21)).tolist()
            plt.imshow(rgb_image)  # plot the image for matplotlib
            current_axis = plt.gca()
            current_axis.set_axis_off()
            current_axis.xaxis.set_major_locator(NullLocator())
            current_axis.yaxis.set_major_locator(NullLocator())

            # scale each detection back up to the image
            for i in range(detections.size(1)):
                j = 0
                while detections[0, i, j, 0] >= 0.6:
                    score = detections[0, i, j, 0]
                    label_name = VOC_CLASSES[i - 1]
                    display_txt = '%s: %.2f' % (label_name, score)
                    pt = (detections[0, i, j, 1:] * scale).cpu().numpy()
                    coords = (pt[0], pt[1]), pt[2] - pt[0] + 1, pt[3] - pt[1] + 1
                    color = colors[i]
                    current_axis.add_patch(plt.Rectangle(*coords, fill=False, edgecolor=color, linewidth=2))
                    current_axis.text(pt[0], pt[1], display_txt, bbox={'facecolor': color, 'alpha': 0.5})
                    j += 1
            plt.axis('off')
            plt.tight_layout()
            save_name = input_data.split('/')[-1]
            save_name = save_name.split('.')
            save_name = '.'.join(save_name[:-1]) + "_prediction." + save_name[-1]
            plt.savefig(os.path.join(output_file_path, save_name), bbox_inches='tight', pad_inches=0)
            plt.clf()

        return results


def base_transform(image, size, mean):
    x = cv2.resize(image, (size, size)).astype(np.float32)
    x -= mean
    x = x.astype(np.float32)
    return x



from distutils.core import setup
from setuptools import find_packages

setup(
    name='autokeras',
    packages=find_packages(exclude=('tests',)),
    install_requires=['scipy==1.2.0',
                      'torch==1.0.1',
                      'torchvision==0.2.1',
                      'numpy==1.15.4',
                      'keras==2.2.4',
                      'scikit-learn==0.20.2',
                      'scikit-image==0.14.2',
                      'tqdm==4.29.1',
                      'tensorflow==1.12.0',
                      'imageio==2.4.1',
                      'requests==2.21.0',
                      'lightgbm==2.2.3',
                      'pandas==0.23.4',
                      'librosa==0.6.2',
                      'numba',
                      'inflect',
                      'unidecode',
                      'nltk==3.3',
                      'lws==1.2',
                      'opencv-python==4.0.0.21',
                      'boto3'],
    version='0.3.7',
    description='AutoML for deep learning',
    author='DATA Lab at Texas A&M University',
    author_email='jhfjhfj1@gmail.com',
    url='http://autokeras.com',
    download_url='https://github.com/jhfjhfj1/autokeras/archive/0.3.7.tar.gz',
    keywords=['AutoML', 'keras'],
    classifiers=[]
)



from autokeras.pretrained import FaceDetector
import os
import urllib.request


def test_face_detection():
    img_file, out_file = 'test.jpg', 'output.jpg'
    urllib.request.urlretrieve('https://raw.githubusercontent.com/kuaikuaikim/DFace/master/test.jpg', img_file)
    if os.path.exists(out_file):
        os.remove(out_file)
    face_detection = FaceDetector()
    bboxs1, landmarks1 = face_detection.predict(img_file, out_file)
    assert os.path.exists(out_file)
    bboxs2, landmarks2 = face_detection.predict(img_file)
    assert bboxs1.shape == bboxs2.shape == (11, 5) and landmarks1.shape == landmarks2.shape == (11, 10)
    os.remove(img_file)
    os.remove(out_file)



import lws
import os

import librosa
import numpy as np
import torch
from scipy import signal

from autokeras.constant import Constant
from autokeras.pretrained.base import Pretrained
from autokeras.pretrained.voice_generator.deepvoice3_pytorch import frontend, builder
from autokeras.utils import temp_path_generator, ensure_dir, get_device, download_file_from_google_drive


# NOTE: If you want full control for model architecture. please take a look
# at the code and change whatever you want. Some hyper parameters are hardcoded.

# Default hyperparameters:

class Hparams:
    name = "deepvoice3"

    # Text:
    # [en  jp]
    frontend = 'en'

    # Replace words to its pronunciation with fixed probability.
    # e.g.  'hello' to 'HH AH0 L OW1'
    # [en  jp]
    # en: Word -> pronunciation using CMUDict
    # jp: Word -> pronounciation usnig MeCab
    # [0 ~ 1.0]: 0 means no replacement happens.
    replace_pronunciation_prob = 0.5

    # Convenient model builder
    # Definitions can be found at deepvoice3_pytorch/builder.py
    # deepvoice3: DeepVoice3 https://arxiv.org/abs/1710.07654
    builder = "deepvoice3"

    # Must be configured depends on the dataset and model you use
    n_speakers = 1
    speaker_embed_dim = 16

    # Audio:
    num_mels = 80
    fmin = 125
    fmax = 7600
    fft_size = 1024
    hop_size = 256
    sample_rate = 22050
    preemphasis = 0.97
    min_level_db = -100
    ref_level_db = 20
    # whether to rescale waveform or not.
    # Let x is an input waveform  rescaled waveform y is given by:
    # y = x / np.abs(x).max() * rescaling_max
    rescaling = False
    rescaling_max = 0.999
    # mel-spectrogram is normalized to [0  1] for each utterance and clipping may
    # happen depends on min_level_db and ref_level_db  causing clipping noise.
    # If False  assertion is added to ensure no clipping happens.
    allow_clipping_in_normalization = True

    # Model:
    downsample_step = 4  # must be 4 when builder="nyanko"
    outputs_per_step = 1  # must be 1 when builder="nyanko"
    embedding_weight_std = 0.1
    speaker_embedding_weight_std = 0.01
    padding_idx = 0
    # Maximum number of input text length
    # try setting larger value if you want to give very long text input
    max_positions = 512
    dropout = 1 - 0.95
    kernel_size = 3
    text_embed_dim = 128
    encoder_channels = 256
    decoder_channels = 256
    # Note: large converter channels requires significant computational cost
    converter_channels = 256
    query_position_rate = 1.0
    # can be computed by `compute_timestamp_ratio.py`.
    key_position_rate = 1.385  # 2.37 for jsut
    key_projection = False
    value_projection = False
    use_memory_mask = True
    trainable_positional_encodings = False
    freeze_embedding = False
    # If True  use decoder's internal representation for postnet inputs
    # otherwise use mel-spectrogram.
    use_decoder_state_for_postnet_input = True

    # Data loader
    pin_memory = True
    num_workers = 2  # Set it to 1 when in Windows (MemoryError  THAllocator.c 0x5)

    # Loss
    masked_loss_weight = 0.5  # (1-w)*loss + w * masked_loss
    priority_freq = 3000  # heuristic: priotrize [0 ~ priotiry_freq] for linear loss
    priority_freq_weight = 0.0  # (1-w)*linear_loss + w*priority_linear_loss
    # https://arxiv.org/pdf/1710.08969.pdf
    # Adding the divergence to the loss stabilizes training  expecially for
    # very deep (> 10 layers) networks.
    # Binary div loss seems has approx 10x scale compared to L1 loss  so I choose 0.1.
    binary_divergence_weight = 0.1  # set 0 to disable
    use_guided_attention = True
    guided_attention_sigma = 0.2

    # Training:
    batch_size = 16
    adam_beta1 = 0.5
    adam_beta2 = 0.9
    adam_eps = 1e-6
    amsgrad = False
    initial_learning_rate = 5e-4  # 0.001
    lr_schedule = "noam_learning_rate_decay"
    lr_schedule_kwargs = {}
    nepochs = 2000
    weight_decay = 0.0
    clip_thresh = 0.1

    # Save
    checkpoint_interval = 10000
    eval_interval = 10000
    save_optimizer_state = True

    # Eval:
    # this can be list for multple layers of attention
    # e.g.  [True  False  False  False  True]
    force_monotonic_attention = True
    # Attention constraint for incremental decoding
    window_ahead = 3
    # 0 tends to prevent word repretetion  but sometime causes skip words
    window_backward = 1
    power = 1.4  # Power to raise magnitudes to prior to phase retrieval

    # GC:
    # Forced garbage collection probability
    # Use only when MemoryError continues in Windows (Disabled by default)
    # gc_probability = 0.001

    # json_meta mode only
    # 0: "use all"
    # 1: "ignore only unmatched_alignment"
    # 2: "fully ignore recognition"
    ignore_recognition_level = 2
    # when dealing with non-dedicated speech dataset(e.g. movie excerpts)  setting min_text above 15 is desirable.
    # Can be adjusted by dataset.
    min_text = 20
    # if true  data without phoneme alignment file(.lab) will be ignored
    process_only_htk_aligned = False


fs = Hparams.sample_rate
global_step = 0
global_epoch = 0


def build_model():
    model = getattr(builder, Hparams.builder)(
        n_speakers=Hparams.n_speakers,
        speaker_embed_dim=Hparams.speaker_embed_dim,
        n_vocab=frontend.n_vocab,
        embed_dim=Hparams.text_embed_dim,
        mel_dim=Hparams.num_mels,
        linear_dim=Hparams.fft_size // 2 + 1,
        r=Hparams.outputs_per_step,
        padding_idx=Hparams.padding_idx,
        dropout=Hparams.dropout,
        kernel_size=Hparams.kernel_size,
        encoder_channels=Hparams.encoder_channels,
        decoder_channels=Hparams.decoder_channels,
        converter_channels=Hparams.converter_channels,
        use_memory_mask=Hparams.use_memory_mask,
        trainable_positional_encodings=Hparams.trainable_positional_encodings,
        force_monotonic_attention=Hparams.force_monotonic_attention,
        use_decoder_state_for_postnet_input=Hparams.use_decoder_state_for_postnet_input,
        max_positions=Hparams.max_positions,
        freeze_embedding=Hparams.freeze_embedding,
        window_ahead=Hparams.window_ahead,
        window_backward=Hparams.window_backward
    )
    return model


def load_checkpoint(path, model, device):
    global global_step
    global global_epoch

    print("Load checkpoint from: {}".format(path))
    if device.startswith("cuda"):
        checkpoint = torch.load(path)
    else:
        checkpoint = torch.load(path, map_location=lambda storage, loc: storage)
    model.load_state_dict(checkpoint["state_dict"])
    global_step = checkpoint["global_step"]
    global_epoch = checkpoint["global_epoch"]

    return model


def inv_preemphasis(x, coef=Hparams.preemphasis):
    """Inverse operation of pre-emphasis

    Args:
        x (1d-array): Input signal.
        coef (float): Pre-emphasis coefficient.

    Returns:
        array: Output filtered signal.

    See also:
        :func:`preemphasis`
    """
    b = np.array([1.], x.dtype)
    a = np.array([1., -coef], x.dtype)
    return signal.lfilter(b, a, x)


def inv_spectrogram(spectrogram):
    """Converts spectrogram to waveform using librosa"""
    S = _db_to_amp(_denormalize(spectrogram) + Hparams.ref_level_db)  # Convert back to linear
    processor = _lws_processor()
    D = processor.run_lws(S.astype(np.float64).T ** Hparams.power)
    y = processor.istft(D).astype(np.float32)
    return inv_preemphasis(y)


def _lws_processor():
    return lws.lws(Hparams.fft_size, Hparams.hop_size, mode="speech")


_mel_basis = None


def _db_to_amp(x):
    return np.power(10.0, x * 0.05)


def _denormalize(S):
    return (np.clip(S, 0, 1) * -Hparams.min_level_db) + Hparams.min_level_db


class VoiceGenerator(Pretrained):
    def __init__(self, model_path=None, overwrite=False):
        super(VoiceGenerator, self).__init__()
        self.model_path = model_path if model_path is not None else temp_path_generator()
        ensure_dir(self.model_path)
        self.checkpoint_path = os.path.join(self.model_path, Constant.PRE_TRAIN_VOICE_GENERATOR_MODEL_NAME)
        self.sample_rate = 0
        self.hop_length = 0
        self.overwrite = overwrite
        self.device = get_device()
        self.load()

    def load(self):
        self._maybe_download()
        self.sample_rate = Hparams.sample_rate
        self.hop_length = Hparams.hop_size
        model = build_model()

        self.model = load_checkpoint(self.checkpoint_path, model, self.device)
        self.model.to(self.device)

    def _maybe_download(self):
        # For files in dropbox or google drive, cannot directly use request to download
        # This can be changed directly use download_file method when the file is stored in server
        if not os.path.exists(self.checkpoint_path) or self.overwrite:
            checkpoint_google_id = Constant.PRE_TRAIN_VOICE_GENERATOR_MODEL_GOOGLE_DRIVE_ID
            download_file_from_google_drive(file_id=checkpoint_google_id, dest_path=self.checkpoint_path,
                                            overwrite=self.overwrite)

    def generate(self, text, path=None):
        waveform, alignment, spectrogram, mel = self.tts(text)
        if path is None:
            path = Constant.PRE_TRAIN_VOICE_GENERATOR_SAVE_FILE_DEFAULT_NAME
        librosa.output.write_wav(path, waveform, self.sample_rate)

    def predict(self, x_predict):
        pass

    def tts(self, text, p=0, speaker_id=None, fast=True):
        """Convert text to speech waveform given a deepvoice3 model.

        Args:
            text (str) : Input text to be synthesized
            p (float) : Replace word to pronounciation if p > 0. Default is 0.
        """
        self.model.eval()
        if fast:
            self.model.make_generation_fast_()

        sequence = np.array(frontend.text_to_sequence(text, p=p))
        sequence = torch.from_numpy(sequence).unsqueeze(0).long().to(self.device)
        text_positions = torch.arange(1, sequence.size(-1) + 1).unsqueeze(0).long().to(self.device)
        speaker_ids = None if speaker_id is None else torch.LongTensor([speaker_id]).to(self.device)

        # Greedy decoding
        with torch.no_grad():
            mel_outputs, linear_outputs, alignments, done = self.model(
                sequence, text_positions=text_positions, speaker_ids=speaker_ids)

        linear_output = linear_outputs[0].cpu().data.numpy()
        spectrogram = _denormalize(linear_output)
        alignment = alignments[0].cpu().data.numpy()
        mel = mel_outputs[0].cpu().data.numpy()
        mel = _denormalize(mel)

        # Predicted audio signal
        waveform = inv_spectrogram(linear_output.T)

        return waveform, alignment, spectrogram, mel
